# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

from ansible.module_utils.basic import AnsibleModule
from ansible.modules import mount_facts
from units.modules.utils import (
    AnsibleExitJson,
    AnsibleFailJson,
    exit_json,
    fail_json,
)
from units.modules.mount_facts_data import aix7_2, freebsd14_1, openbsd6_4, rhel9_4, solaris11_2

from dataclasses import astuple
from unittest.mock import MagicMock

import pytest
import time


@pytest.fixture
def get_mock_module(set_module_args):
    def get_mock_module(invocation):
        set_module_args(invocation)
        module = AnsibleModule(
            argument_spec=mount_facts.get_argument_spec(),
            supports_check_mode=True,
        )
        return module

    yield get_mock_module


def mock_callable(return_value=None):
    def func(*args, **kwargs):
        return return_value

    return func


@pytest.fixture
def set_file_content(monkeypatch):
    def _set_file_content(data):
        def mock_get_file_content(filename, default=None):
            if filename in data:
                return data[filename]
            return default
        monkeypatch.setattr(mount_facts, "get_file_content", mock_get_file_content)
    return _set_file_content


def test_invocation(monkeypatch, set_file_content, set_module_args):
    set_file_content({})
    set_module_args({})
    monkeypatch.setattr(mount_facts, "run_mount_bin", mock_callable(return_value=""))
    monkeypatch.setattr(AnsibleModule, "exit_json", exit_json)
    with pytest.raises(AnsibleExitJson, match="{'ansible_facts': {'mount_points': {}, 'aggregate_mounts': \\[\\]}}"):
        mount_facts.main()


def test_invalid_timeout(monkeypatch, set_module_args):
    set_module_args({"timeout": 0})
    monkeypatch.setattr(AnsibleModule, "fail_json", fail_json)
    with pytest.raises(AnsibleFailJson, match="argument 'timeout' must be a positive number or null"):
        mount_facts.main()


@pytest.mark.parametrize("invalid_binary, error", [
    ("ansible_test_missing_binary", 'Failed to find required executable "ansible_test_missing_binary" in paths: .*'),
    ("false", "Failed to execute .*false: Command '[^']*false' returned non-zero exit status 1"),
    (["one", "two"], "argument 'mount_binary' must be a string or null, not \\['one', 'two'\\]"),
])
def test_invalid_mount_binary(monkeypatch, set_file_content, invalid_binary, error, set_module_args):
    set_file_content({})
    set_module_args({"mount_binary": invalid_binary})
    monkeypatch.setattr(AnsibleModule, "fail_json", fail_json)
    with pytest.raises(AnsibleFailJson, match=error):
        mount_facts.main()


def test_invalid_source(monkeypatch, set_module_args):
    set_module_args({"sources": [""]})
    monkeypatch.setattr(AnsibleModule, "fail_json", fail_json)
    with pytest.raises(AnsibleFailJson, match="sources contains an empty string"):
        mount_facts.main()


def test_duplicate_source(monkeypatch, set_file_content, get_mock_module):
    set_file_content({"/etc/fstab": rhel9_4.fstab})
    mock_warn = MagicMock()
    monkeypatch.setattr(AnsibleModule, "warn", mock_warn)

    user_sources = ["static", "/etc/fstab"]
    resolved_sources = ["/etc/fstab", "/etc/vfstab", "/etc/filesystems", "/etc/fstab"]
    expected_warning = "mount_facts option 'sources' contains duplicate entries, repeat sources will be ignored:"

    module = get_mock_module({"sources": user_sources})
    list(mount_facts.gen_mounts_by_source(module))
    assert mock_warn.called
    assert len(mock_warn.call_args_list) == 1
    assert mock_warn.call_args_list[0].args == (f"{expected_warning} {resolved_sources}",)


@pytest.mark.parametrize("function, data, expected_result_data", [
    ("gen_fstab_entries", rhel9_4.fstab, rhel9_4.fstab_parsed),
    ("gen_fstab_entries", rhel9_4.mtab, rhel9_4.mtab_parsed),
    ("gen_fstab_entries", freebsd14_1.fstab, freebsd14_1.fstab_parsed),
    ("gen_vfstab_entries", solaris11_2.vfstab, solaris11_2.vfstab_parsed),
    ("gen_mnttab_entries", solaris11_2.mnttab, solaris11_2.mnttab_parsed),
    ("gen_aix_filesystems_entries", aix7_2.filesystems, aix7_2.filesystems_parsed),
])
def test_list_mounts(function, data, expected_result_data):
    function = getattr(mount_facts, function)
    result = [astuple(mount_info) for mount_info in function(data.splitlines())]
    assert result == [(_mnt, _line, _info) for _src, _mnt, _line, _info in expected_result_data]


def test_etc_filesystems_linux(set_file_content, get_mock_module):
    not_aix_data = (
        "ext4\next3\next2\nnodev proc\nnodev devpts\niso9770\nvfat\nhfs\nhfsplus\n*"
    )
    set_file_content({"/etc/filesystems": not_aix_data})
    module = get_mock_module({"sources": ["/etc/filesystems"]})
    assert list(mount_facts.gen_mounts_by_source(module)) == []


@pytest.mark.parametrize("mount_stdout, expected_result_data", [
    (rhel9_4.mount, rhel9_4.mount_parsed),
    (freebsd14_1.mount, freebsd14_1.mount_parsed),
    (openbsd6_4.mount, openbsd6_4.mount_parsed),
    (aix7_2.mount, aix7_2.mount_parsed),
])
def test_parse_mount_bin_stdout(mount_stdout, expected_result_data):
    expected_result = [(_mnt, _line, _info) for _src, _mnt, _line, _info in expected_result_data]
    result = [astuple(mount_info) for mount_info in mount_facts.gen_mounts_from_stdout(mount_stdout)]
    assert result == expected_result


def test_parse_mount_bin_stdout_unknown(monkeypatch, set_file_content, set_module_args):
    set_file_content({})
    set_module_args({})
    monkeypatch.setattr(mount_facts, "run_mount_bin", mock_callable(return_value="nonsense"))
    monkeypatch.setattr(AnsibleModule, "exit_json", exit_json)
    with pytest.raises(AnsibleExitJson, match="{'ansible_facts': {'mount_points': {}, 'aggregate_mounts': \\[\\]}}"):
        mount_facts.main()


rhel_mock_fs = {"/etc/fstab": rhel9_4.fstab, "/etc/mtab": rhel9_4.mtab}
freebsd_mock_fs = {"/etc/fstab": freebsd14_1.fstab}
aix_mock_fs = {"/etc/filesystems": aix7_2.filesystems}
openbsd_mock_fs = {"/etc/fstab": openbsd6_4.fstab}
solaris_mock_fs = {"/etc/mnttab": solaris11_2.mnttab, "/etc/vfstab": solaris11_2.vfstab}


@pytest.mark.parametrize("sources, file_data, mount_data, results", [
    (["static"], rhel_mock_fs, rhel9_4.mount, rhel9_4.fstab_parsed),
    (["/etc/fstab"], rhel_mock_fs, rhel9_4.mount, rhel9_4.fstab_parsed),
    (["dynamic"], rhel_mock_fs, rhel9_4.mount, rhel9_4.mtab_parsed),
    (["/etc/mtab"], rhel_mock_fs, rhel9_4.mount, rhel9_4.mtab_parsed),
    (["all"], rhel_mock_fs, rhel9_4.mount, rhel9_4.mtab_parsed + rhel9_4.fstab_parsed),
    (["mount"], rhel_mock_fs, rhel9_4.mount, rhel9_4.mount_parsed),
    (["all"], freebsd_mock_fs, freebsd14_1.mount, freebsd14_1.fstab_parsed + freebsd14_1.mount_parsed),
    (["static"], freebsd_mock_fs, freebsd14_1.mount, freebsd14_1.fstab_parsed),
    (["dynamic"], freebsd_mock_fs, freebsd14_1.mount, freebsd14_1.mount_parsed),
    (["mount"], freebsd_mock_fs, freebsd14_1.mount, freebsd14_1.mount_parsed),
    (["all"], aix_mock_fs, aix7_2.mount, aix7_2.filesystems_parsed + aix7_2.mount_parsed),
    (["all"], openbsd_mock_fs, openbsd6_4.mount, openbsd6_4.fstab_parsed + openbsd6_4.mount_parsed),
    (["all"], solaris_mock_fs, "", solaris11_2.mnttab_parsed + solaris11_2.vfstab_parsed),
])
def test_gen_mounts_by_sources(monkeypatch, set_file_content, sources, file_data, mount_data, results, get_mock_module):
    set_file_content(file_data)
    mock_run_mount = mock_callable(return_value=mount_data)
    monkeypatch.setattr(mount_facts, "run_mount_bin", mock_run_mount)
    module = get_mock_module({"sources": sources})
    actual_results = list(mount_facts.gen_mounts_by_source(module))
    assert actual_results == results


@pytest.mark.parametrize("on_timeout, should_warn, should_raise", [
    (None, False, True),
    ("warn", True, False),
    ("ignore", False, False),
])
def test_gen_mounts_by_source_timeout(monkeypatch, set_file_content, on_timeout, should_warn, should_raise, get_mock_module):
    set_file_content({})
    monkeypatch.setattr(AnsibleModule, "fail_json", fail_json)
    mock_warn = MagicMock()
    monkeypatch.setattr(AnsibleModule, "warn", mock_warn)

    # hack to simulate a hang (module entrypoint requires >= 1 or None)
    params = {"timeout": 0}
    if on_timeout:
        params["on_timeout"] = on_timeout

    module = get_mock_module(params)
    if should_raise:
        with pytest.raises(AnsibleFailJson, match="Command '[^']*mount' timed out after 0.0 seconds"):
            list(mount_facts.gen_mounts_by_source(module))
    else:
        assert list(mount_facts.gen_mounts_by_source(module)) == []
    assert mock_warn.called == should_warn


def test_get_mount_facts(monkeypatch, set_file_content, get_mock_module):
    set_file_content({"/etc/mtab": rhel9_4.mtab, "/etc/fstab": rhel9_4.fstab})
    monkeypatch.setattr(mount_facts, "get_mount_size", mock_callable(return_value=None))
    monkeypatch.setattr(mount_facts, "get_partition_uuid", mock_callable(return_value=None))
    module = get_mock_module({})
    assert len(rhel9_4.fstab_parsed) == 3
    assert len(rhel9_4.mtab_parsed) == 4
    assert len(mount_facts.get_mount_facts(module)) == 7


@pytest.mark.parametrize("filter_name, filter_value, source, mount_info", [
    ("devices", "proc", rhel9_4.mtab_parsed[0][2], rhel9_4.mtab_parsed[0][-1]),
    ("fstypes", "proc", rhel9_4.mtab_parsed[0][2], rhel9_4.mtab_parsed[0][-1]),
])
def test_get_mounts_facts_filtering(monkeypatch, set_file_content, filter_name, filter_value, source, mount_info, get_mock_module):
    set_file_content({"/etc/mtab": rhel9_4.mtab})
    monkeypatch.setattr(mount_facts, "get_mount_size", mock_callable(return_value=None))
    monkeypatch.setattr(mount_facts, "get_partition_uuid", mock_callable(return_value=None))
    module = get_mock_module({filter_name: [filter_value]})
    results = mount_facts.get_mount_facts(module)
    expected_context = {"source": "/etc/mtab", "source_data": source}
    assert len(results) == 1
    assert results[0]["ansible_context"] == expected_context
    assert results[0] == dict(ansible_context=expected_context, uuid=None, **mount_info)


def test_get_mounts_size(monkeypatch, set_file_content, get_mock_module):
    def mock_get_mount_size(*args, **kwargs):
        return {
            "block_available": 3242510,
            "block_size": 4096,
            "block_total": 3789825,
            "block_used": 547315,
            "inode_available": 1875503,
            "inode_total": 1966080,
            "size_available": 13281320960,
            "size_total": 15523123200,
        }

    set_file_content({"/etc/mtab": rhel9_4.mtab})
    monkeypatch.setattr(mount_facts, "get_mount_size", mock_get_mount_size)
    monkeypatch.setattr(mount_facts, "get_partition_uuid", mock_callable(return_value=None))
    module = get_mock_module({})
    result = mount_facts.get_mount_facts(module)
    assert len(result) == len(rhel9_4.mtab_parsed)
    expected_results = setup_parsed_sources_with_context(rhel9_4.mtab_parsed)
    assert result == [dict(**result, **mock_get_mount_size()) for result in expected_results]


@pytest.mark.parametrize("on_timeout, should_warn, should_raise", [
    (None, False, True),
    ("error", False, True),
    ("warn", True, False),
    ("ignore", False, False),
])
def test_get_mount_size_timeout(monkeypatch, set_file_content, on_timeout, should_warn, should_raise, get_mock_module):
    set_file_content({"/etc/mtab": rhel9_4.mtab})
    monkeypatch.setattr(AnsibleModule, "fail_json", fail_json)
    mock_warn = MagicMock()
    monkeypatch.setattr(AnsibleModule, "warn", mock_warn)
    monkeypatch.setattr(mount_facts, "get_partition_uuid", mock_callable(return_value=None))

    params = {"timeout": 0.1, "sources": ["dynamic"], "mount_binary": "mount"}
    if on_timeout:
        params["on_timeout"] = on_timeout

    module = get_mock_module(params)

    def mock_slow_function(*args, **kwargs):
        time.sleep(0.15)
        raise Exception("Timeout failed")

    monkeypatch.setattr(mount_facts, "get_mount_size", mock_slow_function)
    match = r"Timed out getting mount size for mount /proc \(type proc\) after 0.1 seconds"

    if should_raise:
        with pytest.raises(AnsibleFailJson, match=match):
            mount_facts.get_mount_facts(module)
    else:
        results = mount_facts.get_mount_facts(module)
        assert len(results) == len(rhel9_4.mtab_parsed)
        assert results == setup_parsed_sources_with_context(rhel9_4.mtab_parsed)
    assert mock_warn.called == should_warn


def setup_missing_uuid(monkeypatch, module):
    monkeypatch.setattr(module, "get_bin_path", mock_callable(return_value="fake_bin"))
    monkeypatch.setattr(mount_facts.subprocess, "check_output", mock_callable(return_value=""))
    monkeypatch.setattr(mount_facts.os, "listdir", MagicMock(side_effect=FileNotFoundError))


def setup_udevadm_fallback(monkeypatch):
    mock_udevadm_snippet = (
        "DEVPATH=/devices/virtual/block/dm-0\n"
        "DEVNAME=/dev/dm-0\n"
        "DEVTYPE=disk\n"
        "ID_FS_UUID=d37b5483-304d-471d-913e-4bb77856d90f\n"
    )
    monkeypatch.setattr(mount_facts.subprocess, "check_output", mock_callable(return_value=mock_udevadm_snippet))


def setup_run_lsblk_fallback(monkeypatch):
    mount_facts.run_lsblk.cache_clear()
    mock_lsblk_snippet = (
        "/dev/zram0\n"
        "/dev/mapper/luks-2f2610e3-56b3-4131-ae3e-caf9aa535b73 d37b5483-304d-471d-913e-4bb77856d90f\n"
        "/dev/nvme0n1\n"
        "/dev/nvme0n1p1                                        B521-06FD\n"
    )
    monkeypatch.setattr(mount_facts.subprocess, "check_output", mock_callable(return_value=mock_lsblk_snippet))


def setup_list_uuids_linux_preferred(monkeypatch):
    mount_facts.list_uuids_linux.cache_clear()

    def mock_realpath(path):
        if path.endswith("d37b5483-304d-471d-913e-4bb77856d90f"):
            return "/dev/mapper/luks-2f2610e3-56b3-4131-ae3e-caf9aa535b73"
        return 'miss'

    monkeypatch.setattr(mount_facts.os.path, "realpath", mock_realpath)
    monkeypatch.setattr(mount_facts.os, "listdir", mock_callable(return_value=["B521-06FD", "d37b5483-304d-471d-913e-4bb77856d90f"]))


def test_get_partition_uuid(monkeypatch, get_mock_module):
    module = get_mock_module({})

    # Test missing UUID
    setup_missing_uuid(monkeypatch, module)
    assert mount_facts.get_partition_uuid(module, "ansible-test-fake-dev") is None

    # Test udevadm (legacy fallback)
    setup_udevadm_fallback(monkeypatch)
    assert mount_facts.get_partition_uuid(module, "ansible-test-udevadm") == "d37b5483-304d-471d-913e-4bb77856d90f"

    # Test run_lsblk fallback
    setup_run_lsblk_fallback(monkeypatch)
    assert mount_facts.get_partition_uuid(module, "/dev/nvme0n1p1") == "B521-06FD"

    # Test list_uuids_linux (preferred)
    setup_list_uuids_linux_preferred(monkeypatch)
    assert mount_facts.get_partition_uuid(module, "/dev/mapper/luks-2f2610e3-56b3-4131-ae3e-caf9aa535b73") == "d37b5483-304d-471d-913e-4bb77856d90f"


def setup_parsed_sources_with_context(parsed_sources):
    mounts = []
    for source, mount, line, info in parsed_sources:
        mount_point_result = dict(ansible_context={"source": source, "source_data": line}, uuid=None, **info)
        mounts.append(mount_point_result)
    return mounts


def test_handle_deduplication(monkeypatch, get_mock_module):
    unformatted_mounts = [
        {"mount": "/mnt/mount1", "device": "foo", "ansible_context": {"source": "/proc/mounts"}},
        {"mount": "/mnt/mount2", "ansible_context": {"source": "/proc/mounts"}},
        {"mount": "/mnt/mount1", "device": "qux", "ansible_context": {"source": "/etc/fstab"}},
    ]

    mock_warn = MagicMock()
    monkeypatch.setattr(AnsibleModule, "warn", mock_warn)

    module = get_mock_module({})
    mount_points, aggregate_mounts = mount_facts.handle_deduplication(module, unformatted_mounts)
    assert len(mount_points.keys()) == 2
    assert mount_points["/mnt/mount1"]["device"] == "foo"
    assert aggregate_mounts == []
    assert not mock_warn.called


@pytest.mark.parametrize("include_aggregate_mounts, warn_expected", [
    (None, True),
    (False, False),
    (True, False),
])
def test_handle_deduplication_warning(monkeypatch, include_aggregate_mounts, warn_expected, get_mock_module):
    unformatted_mounts = [
        {"mount": "/mnt/mount1", "device": "foo", "ansible_context": {"source": "/proc/mounts"}},
        {"mount": "/mnt/mount1", "device": "bar", "ansible_context": {"source": "/proc/mounts"}},
        {"mount": "/mnt/mount2", "ansible_context": {"source": "/proc/mounts"}},
        {"mount": "/mnt/mount1", "device": "qux", "ansible_context": {"source": "/etc/fstab"}},
    ]

    mock_warn = MagicMock()
    monkeypatch.setattr(AnsibleModule, "warn", mock_warn)

    module = get_mock_module({"include_aggregate_mounts": include_aggregate_mounts})
    mount_points, aggregate_mounts = mount_facts.handle_deduplication(module, unformatted_mounts)

    assert aggregate_mounts == [] if not include_aggregate_mounts else unformatted_mounts
    assert mock_warn.called == warn_expected



# (c) 2019 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# CAUTION: This implementation of the collection loader is used by ansible-test.
#          Because of this, it must be compatible with all Python versions supported on the controller or remote.

from __future__ import annotations

import itertools
import os
import os.path
import pathlib
import re
import sys

from contextlib import contextmanager
from importlib import import_module, reload as reload_module
from importlib.machinery import FileFinder
from importlib.util import find_spec, spec_from_loader
from keyword import iskeyword
from types import ModuleType

# DO NOT add new non-stdlib import deps here, this loader is used by external tools (eg ansible-test import sanity)
# that only allow stdlib and module_utils
from . import _to_bytes, _to_text
from ._collection_config import AnsibleCollectionConfig

try:
    try:
        # Available on Python >= 3.11
        # We ignore the import error that will trigger when running mypy with
        # older Python versions.
        from importlib.resources.abc import TraversableResources  # type: ignore[import]
    except ImportError:
        # Used with Python 3.9 and 3.10 only
        # This member is still available as an alias up until Python 3.14 but
        # is deprecated as of Python 3.12.
        from importlib.abc import TraversableResources  # deprecated: description='TraversableResources move' python_version='3.10'
except ImportError:
    # Python < 3.9
    # deprecated: description='TraversableResources fallback' python_version='3.8'
    TraversableResources = object  # type: ignore[assignment,misc]

# NB: this supports import sanity test providing a different impl
try:
    from ._collection_meta import _meta_yml_to_dict
except ImportError:
    _meta_yml_to_dict = None

PB_EXTENSIONS = ('.yml', '.yaml')
SYNTHETIC_PACKAGE_NAME = '<ansible_synthetic_collection_package>'


class _AnsibleNSTraversable:
    """Class that implements the ``importlib.resources.abc.Traversable``
    interface for the following ``ansible_collections`` namespace packages::

    * ``ansible_collections``
    * ``ansible_collections.<namespace>``

    These namespace packages operate differently from a normal Python
    namespace package, in that the same namespace can be distributed across
    multiple directories on the filesystem and still function as a single
    namespace, such as::

    * ``/usr/share/ansible/collections/ansible_collections/ansible/posix/``
    * ``/home/user/.ansible/collections/ansible_collections/ansible/windows/``

    This class will mimic the behavior of various ``pathlib.Path`` methods,
    by combining the results of multiple root paths into the output.

    This class does not do anything to remove duplicate collections from the
    list, so when traversing either namespace patterns supported by this class,
    it is possible to have the same collection located in multiple root paths,
    but precedence rules only use one. When iterating or traversing these
    package roots, there is the potential to see the same collection in
    multiple places without indication of which would be used. In such a
    circumstance, it is best to then call ``importlib.resources.files`` for an
    individual collection package rather than continuing to traverse from the
    namespace package.

    Several methods will raise ``NotImplementedError`` as they do not make
    sense for these namespace packages.
    """
    def __init__(self, *paths):
        self._paths = [pathlib.Path(p) for p in paths]

    def __repr__(self):
        return "_AnsibleNSTraversable('%s')" % "', '".join(map(_to_text, self._paths))

    def iterdir(self):
        return itertools.chain.from_iterable(p.iterdir() for p in self._paths if p.is_dir())

    def is_dir(self):
        return any(p.is_dir() for p in self._paths)

    def is_file(self):
        return False

    def glob(self, pattern):
        return itertools.chain.from_iterable(p.glob(pattern) for p in self._paths if p.is_dir())

    def _not_implemented(self, *args, **kwargs):
        raise NotImplementedError('not usable on namespaces')

    joinpath = __truediv__ = read_bytes = read_text = _not_implemented


class _AnsibleTraversableResources(TraversableResources):
    """Implements ``importlib.resources.abc.TraversableResources`` for the
    collection Python loaders.

    The result of ``files`` will depend on whether a particular collection, or
    a sub package of a collection was referenced, as opposed to
    ``ansible_collections`` or a particular namespace. For a collection and
    its subpackages, a ``pathlib.Path`` instance will be returned, whereas
    for the higher level namespace packages, ``_AnsibleNSTraversable``
    will be returned.
    """
    def __init__(self, package, loader):
        self._package = package
        self._loader = loader

    def _get_name(self, package):
        try:
            # spec
            return package.name
        except AttributeError:
            # module
            return package.__name__

    def _get_package(self, package):
        try:
            # spec
            return package.__parent__
        except AttributeError:
            # module
            return package.__package__

    def _get_path(self, package):
        try:
            # spec
            return package.origin
        except AttributeError:
            # module
            return package.__file__

    def _is_ansible_ns_package(self, package):
        origin = getattr(package, 'origin', None)
        if not origin:
            return False

        if origin == SYNTHETIC_PACKAGE_NAME:
            return True

        module_filename = os.path.basename(origin)
        return module_filename in {'__synthetic__', '__init__.py'}

    def _ensure_package(self, package):
        if self._is_ansible_ns_package(package):
            # Short circuit our loaders
            return
        if self._get_package(package) != package.__name__:
            raise TypeError('%r is not a package' % package.__name__)

    def files(self):
        package = self._package
        parts = package.split('.')
        is_ns = parts[0] == 'ansible_collections' and len(parts) < 3

        if isinstance(package, str):
            if is_ns:
                # Don't use ``spec_from_loader`` here, because that will point
                # to exactly 1 location for a namespace. Use ``find_spec``
                # to get a list of all locations for the namespace
                package = find_spec(package)
            else:
                package = spec_from_loader(package, self._loader)
        elif not isinstance(package, ModuleType):
            raise TypeError('Expected string or module, got %r' % package.__class__.__name__)

        self._ensure_package(package)
        if is_ns:
            return _AnsibleNSTraversable(*package.submodule_search_locations)
        return pathlib.Path(self._get_path(package)).parent


class _AnsibleCollectionFinder:
    def __init__(self, paths=None, scan_sys_paths=True):
        # TODO: accept metadata loader override
        self._ansible_pkg_path = _to_text(os.path.dirname(_to_bytes(sys.modules['ansible'].__file__)))

        if isinstance(paths, str):
            paths = [paths]
        elif paths is None:
            paths = []

        # expand any placeholders in configured paths
        paths = [os.path.expanduser(_to_text(p)) for p in paths]

        # add syspaths if needed
        if scan_sys_paths:
            paths.extend(sys.path)

        good_paths = []
        # expand any placeholders in configured paths
        for p in paths:

            # ensure we always have ansible_collections
            if os.path.basename(p) == 'ansible_collections':
                p = os.path.dirname(p)

            if p not in good_paths and os.path.isdir(_to_bytes(os.path.join(p, 'ansible_collections'))):
                good_paths.append(p)

        self._n_configured_paths = good_paths
        self._n_cached_collection_paths = None
        self._n_cached_collection_qualified_paths = None

        self._n_playbook_paths = []

    @classmethod
    def _find_existing_finder(cls) -> _AnsibleCollectionFinder | None:
        for finder in sys.meta_path:
            if isinstance(finder, _AnsibleCollectionFinder):
                return finder

        return None

    @classmethod
    def _remove(cls):
        for mps in sys.meta_path:
            if isinstance(mps, _AnsibleCollectionFinder):
                sys.meta_path.remove(mps)

        # remove any path hooks that look like ours
        for ph in sys.path_hooks:
            if hasattr(ph, '__self__') and isinstance(ph.__self__, _AnsibleCollectionFinder):
                sys.path_hooks.remove(ph)

        # zap any cached path importer cache entries that might refer to us
        sys.path_importer_cache.clear()

        AnsibleCollectionConfig._collection_finder = None

        # validate via the public property that we really killed it
        if AnsibleCollectionConfig.collection_finder is not None:
            raise AssertionError('_AnsibleCollectionFinder remove did not reset AnsibleCollectionConfig.collection_finder')

    def _install(self):
        self._remove()
        sys.meta_path.insert(0, self)

        sys.path_hooks.insert(0, self._ansible_collection_path_hook)

        AnsibleCollectionConfig.collection_finder = self

    def _ansible_collection_path_hook(self, path):
        path = _to_text(path)
        interesting_paths = self._n_cached_collection_qualified_paths
        if not interesting_paths:
            interesting_paths = []
            for p in self._n_collection_paths:
                if os.path.basename(p) != 'ansible_collections':
                    p = os.path.join(p, 'ansible_collections')

                if p not in interesting_paths:
                    interesting_paths.append(p)

            interesting_paths.insert(0, self._ansible_pkg_path)
            self._n_cached_collection_qualified_paths = interesting_paths

        if any(path.startswith(p) for p in interesting_paths):
            return _AnsiblePathHookFinder(self, path)

        raise ImportError('not interested')

    @property
    def _n_collection_paths(self):
        paths = self._n_cached_collection_paths
        if not paths:
            self._n_cached_collection_paths = paths = self._n_playbook_paths + self._n_configured_paths
        return paths

    def set_playbook_paths(self, playbook_paths):
        if isinstance(playbook_paths, str):
            playbook_paths = [playbook_paths]

        # track visited paths; we have to preserve the dir order as-passed in case there are duplicate collections (first one wins)
        added_paths = set()

        # de-dupe
        self._n_playbook_paths = [os.path.join(_to_text(p), 'collections') for p in playbook_paths if not (p in added_paths or added_paths.add(p))]
        self._n_cached_collection_paths = None
        # HACK: playbook CLI sets this relatively late, so we've already loaded some packages whose paths might depend on this. Fix those up.
        # NB: this should NOT be used for late additions; ideally we'd fix the playbook dir setup earlier in Ansible init
        # to prevent this from occurring
        for pkg in ['ansible_collections', 'ansible_collections.ansible']:
            self._reload_hack(pkg)

    def _reload_hack(self, fullname):
        m = sys.modules.get(fullname)
        if not m:
            return
        reload_module(m)

    def _get_loader(self, fullname, path=None):
        split_name = fullname.split('.')
        toplevel_pkg = split_name[0]
        module_to_find = split_name[-1]
        part_count = len(split_name)

        if toplevel_pkg not in ['ansible', 'ansible_collections']:
            # not interested in anything other than ansible_collections (and limited cases under ansible)
            return None

        # sanity check what we're getting from import, canonicalize path values
        if part_count == 1:
            if path:
                raise ValueError('path should not be specified for top-level packages (trying to find {0})'.format(fullname))
            else:
                # seed the path to the configured collection roots
                path = self._n_collection_paths

        if part_count > 1 and path is None:
            raise ValueError('path must be specified for subpackages (trying to find {0})'.format(fullname))

        if toplevel_pkg == 'ansible':
            # something under the ansible package, delegate to our internal loader in case of redirections
            initialize_loader = _AnsibleInternalRedirectLoader
        elif part_count == 1:
            initialize_loader = _AnsibleCollectionRootPkgLoader
        elif part_count == 2:  # ns pkg eg, ansible_collections, ansible_collections.somens
            initialize_loader = _AnsibleCollectionNSPkgLoader
        elif part_count == 3:  # collection pkg eg, ansible_collections.somens.somecoll
            initialize_loader = _AnsibleCollectionPkgLoader
        else:
            # anything below the collection
            initialize_loader = _AnsibleCollectionLoader

        # NB: actual "find"ing is delegated to the constructors on the various loaders; they'll ImportError if not found
        try:
            return initialize_loader(fullname=fullname, path_list=path)
        except ImportError:
            # TODO: log attempt to load context
            return None

    def find_module(self, fullname, path=None):
        # Figure out what's being asked for, and delegate to a special-purpose loader
        return self._get_loader(fullname, path)

    def find_spec(self, fullname, path, target=None):
        loader = self._get_loader(fullname, path)

        if loader is None:
            return None

        spec = spec_from_loader(fullname, loader)
        if spec is not None and hasattr(loader, '_subpackage_search_paths'):
            spec.submodule_search_locations = loader._subpackage_search_paths
        return spec


# Implements a path_hook finder for iter_modules (since it's only path based). This finder does not need to actually
# function as a finder in most cases, since our meta_path finder is consulted first for *almost* everything, except
# pkgutil.iter_modules, and under py2, pkgutil.get_data if the parent package passed has not been loaded yet.
class _AnsiblePathHookFinder:
    def __init__(self, collection_finder, pathctx):
        # when called from a path_hook, find_module doesn't usually get the path arg, so this provides our context
        self._pathctx = _to_text(pathctx)
        self._collection_finder = collection_finder
        # cache the native FileFinder (take advantage of its filesystem cache for future find/load requests)
        self._file_finder = None

    # class init is fun- this method has a self arg that won't get used
    def _get_filefinder_path_hook(self=None):
        _file_finder_hook = None
        # try to find the FileFinder hook to call for fallback path-based imports in Py3
        _file_finder_hook = [ph for ph in sys.path_hooks if 'FileFinder' in repr(ph)]
        if len(_file_finder_hook) != 1:
            raise Exception('need exactly one FileFinder import hook (found {0})'.format(len(_file_finder_hook)))
        _file_finder_hook = _file_finder_hook[0]

        return _file_finder_hook

    _filefinder_path_hook = _get_filefinder_path_hook()

    def _get_finder(self, fullname):
        split_name = fullname.split('.')
        toplevel_pkg = split_name[0]

        if toplevel_pkg == 'ansible_collections':
            # collections content? delegate to the collection finder
            return self._collection_finder
        else:
            # Something else; we'd normally restrict this to `ansible` descendent modules so that any weird loader
            # behavior that arbitrary Python modules have can be serviced by those loaders. In some dev/test
            # scenarios (eg a venv under a collection) our path_hook signs us up to load non-Ansible things, and
            # it's too late by the time we've reached this point, but also too expensive for the path_hook to figure
            # out what we *shouldn't* be loading with the limited info it has. So we'll just delegate to the
            # normal path-based loader as best we can to service it. This also allows us to take advantage of Python's
            # built-in FS caching and byte-compilation for most things.
            # create or consult our cached file finder for this path
            if not self._file_finder:
                try:
                    self._file_finder = _AnsiblePathHookFinder._filefinder_path_hook(self._pathctx)
                except ImportError:
                    # FUTURE: log at a high logging level? This is normal for things like python36.zip on the path, but
                    # might not be in some other situation...
                    return None

            return self._file_finder

    def find_module(self, fullname, path=None):
        # we ignore the passed in path here- use what we got from the path hook init
        finder = self._get_finder(fullname)

        if finder is None:
            return None
        elif isinstance(finder, FileFinder):
            # this codepath is erroneously used under some cases in py3,
            # and the find_module method on FileFinder does not accept the path arg
            # see https://github.com/pypa/setuptools/pull/2918
            return finder.find_module(fullname)
        else:
            return finder.find_module(fullname, path=[self._pathctx])

    def find_spec(self, fullname, target=None):
        split_name = fullname.split('.')
        toplevel_pkg = split_name[0]

        finder = self._get_finder(fullname)

        if finder is None:
            return None
        elif toplevel_pkg == 'ansible_collections':
            return finder.find_spec(fullname, path=[self._pathctx])
        else:
            return finder.find_spec(fullname)

    def iter_modules(self, prefix):
        # NB: this currently represents only what's on disk, and does not handle package redirection
        return _iter_modules_impl([self._pathctx], prefix)

    def __repr__(self):
        return "{0}(path='{1}')".format(self.__class__.__name__, self._pathctx)


class _AnsibleCollectionPkgLoaderBase:
    _allows_package_code = False

    def __init__(self, fullname, path_list=None):
        self._fullname = fullname
        self._redirect_module = None
        self._split_name = fullname.split('.')
        self._rpart_name = fullname.rpartition('.')
        self._parent_package_name = self._rpart_name[0]  # eg ansible_collections for ansible_collections.somens, '' for toplevel
        self._package_to_load = self._rpart_name[2]  # eg somens for ansible_collections.somens

        self._source_code_path = None
        self._decoded_source = None
        self._compiled_code = None

        self._validate_args()

        self._candidate_paths = self._get_candidate_paths([_to_text(p) for p in path_list])
        self._subpackage_search_paths = self._get_subpackage_search_paths(self._candidate_paths)

        self._validate_final()

    # allow subclasses to validate args and sniff split values before we start digging around
    def _validate_args(self):
        if self._split_name[0] != 'ansible_collections':
            raise ImportError('this loader can only load packages from the ansible_collections package, not {0}'.format(self._fullname))

    # allow subclasses to customize candidate path filtering
    def _get_candidate_paths(self, path_list):
        return [os.path.join(p, self._package_to_load) for p in path_list]

    # allow subclasses to customize finding paths
    def _get_subpackage_search_paths(self, candidate_paths):
        # filter candidate paths for existence (NB: silently ignoring package init code and same-named modules)
        return [p for p in candidate_paths if os.path.isdir(_to_bytes(p))]

    # allow subclasses to customize state validation/manipulation before we return the loader instance
    def _validate_final(self):
        return

    @staticmethod
    @contextmanager
    def _new_or_existing_module(name, **kwargs):
        # handle all-or-nothing sys.modules creation/use-existing/delete-on-exception-if-created behavior
        created_module = False
        module = sys.modules.get(name)
        try:
            if not module:
                module = ModuleType(name)
                created_module = True
                sys.modules[name] = module
            # always override the values passed, except name (allow reference aliasing)
            for attr, value in kwargs.items():
                setattr(module, attr, value)
            yield module
        except Exception:
            if created_module:
                if sys.modules.get(name):
                    sys.modules.pop(name)
            raise

    # basic module/package location support
    # NB: this does not support distributed packages!
    @staticmethod
    def _module_file_from_path(leaf_name, path):
        has_code = True
        package_path = os.path.join(_to_text(path), _to_text(leaf_name))
        module_path = None

        # if the submodule is a package, assemble valid submodule paths, but stop looking for a module
        if os.path.isdir(_to_bytes(package_path)):
            # is there a package init?
            module_path = os.path.join(package_path, '__init__.py')
            if not os.path.isfile(_to_bytes(module_path)):
                module_path = os.path.join(package_path, '__synthetic__')
                has_code = False
        else:
            module_path = package_path + '.py'
            package_path = None
            if not os.path.isfile(_to_bytes(module_path)):
                raise ImportError('{0} not found at {1}'.format(leaf_name, path))

        return module_path, has_code, package_path

    def get_resource_reader(self, fullname):
        return _AnsibleTraversableResources(fullname, self)

    def exec_module(self, module):
        # short-circuit redirect; avoid reinitializing existing modules
        if self._redirect_module:
            return

        # execute the module's code in its namespace
        code_obj = self.get_code(self._fullname)
        if code_obj is not None:  # things like NS packages that can't have code on disk will return None
            exec(code_obj, module.__dict__)

    def create_module(self, spec):
        # short-circuit redirect; we've already imported the redirected module, so just alias it and return it
        if self._redirect_module:
            return self._redirect_module
        else:
            return None

    def load_module(self, fullname):
        # short-circuit redirect; we've already imported the redirected module, so just alias it and return it
        if self._redirect_module:
            sys.modules[self._fullname] = self._redirect_module
            return self._redirect_module

        # we're actually loading a module/package
        module_attrs = dict(
            __loader__=self,
            __file__=self.get_filename(fullname),
            __package__=self._parent_package_name  # sane default for non-packages
        )

        # eg, I am a package
        if self._subpackage_search_paths is not None:  # empty is legal
            module_attrs['__path__'] = self._subpackage_search_paths
            module_attrs['__package__'] = fullname  # per PEP366

        with self._new_or_existing_module(fullname, **module_attrs) as module:
            # execute the module's code in its namespace
            code_obj = self.get_code(fullname)
            if code_obj is not None:  # things like NS packages that can't have code on disk will return None
                exec(code_obj, module.__dict__)

            return module

    def is_package(self, fullname):
        if fullname != self._fullname:
            raise ValueError('this loader cannot answer is_package for {0}, only {1}'.format(fullname, self._fullname))
        return self._subpackage_search_paths is not None

    def get_source(self, fullname):
        if self._decoded_source:
            return self._decoded_source
        if fullname != self._fullname:
            raise ValueError('this loader cannot load source for {0}, only {1}'.format(fullname, self._fullname))
        if not self._source_code_path:
            return None
        # FIXME: what do we want encoding/newline requirements to be?
        self._decoded_source = self.get_data(self._source_code_path)
        return self._decoded_source

    def get_data(self, path):
        if not path:
            raise ValueError('a path must be specified')

        # TODO: ensure we're being asked for a path below something we own
        # TODO: try to handle redirects internally?

        if not path[0] == '/':
            # relative to current package, search package paths if possible (this may not be necessary)
            # candidate_paths = [os.path.join(ssp, path) for ssp in self._subpackage_search_paths]
            raise ValueError('relative resource paths not supported')
        else:
            candidate_paths = [path]

        for p in candidate_paths:
            b_path = _to_bytes(p)
            if os.path.isfile(b_path):
                with open(b_path, 'rb') as fd:
                    return fd.read()
            # HACK: if caller asks for __init__.py and the parent dir exists, return empty string (this keep consistency
            # with "collection subpackages don't require __init__.py" working everywhere with get_data
            elif b_path.endswith(b'__init__.py') and os.path.isdir(os.path.dirname(b_path)):
                return ''

        return None

    def _synthetic_filename(self, fullname):
        return SYNTHETIC_PACKAGE_NAME

    def get_filename(self, fullname):
        if fullname != self._fullname:
            raise ValueError('this loader cannot find files for {0}, only {1}'.format(fullname, self._fullname))

        filename = self._source_code_path

        if not filename and self.is_package(fullname):
            if len(self._subpackage_search_paths) == 1:
                filename = os.path.join(self._subpackage_search_paths[0], '__synthetic__')
            else:
                filename = self._synthetic_filename(fullname)

        return filename

    def get_code(self, fullname):
        if self._compiled_code:
            return self._compiled_code

        # this may or may not be an actual filename, but it's the value we'll use for __file__
        filename = self.get_filename(fullname)
        if not filename:
            filename = '<string>'

        source_code = self.get_source(fullname)

        # for things like synthetic modules that really have no source on disk, don't return a code object at all
        # vs things like an empty package init (which has an empty string source on disk)
        if source_code is None:
            return None

        self._compiled_code = compile(source=source_code, filename=filename, mode='exec', flags=0, dont_inherit=True)

        return self._compiled_code

    def iter_modules(self, prefix):
        return _iter_modules_impl(self._subpackage_search_paths, prefix)

    def __repr__(self):
        return '{0}(path={1})'.format(self.__class__.__name__, self._subpackage_search_paths or self._source_code_path)


class _AnsibleCollectionRootPkgLoader(_AnsibleCollectionPkgLoaderBase):
    def _validate_args(self):
        super(_AnsibleCollectionRootPkgLoader, self)._validate_args()
        if len(self._split_name) != 1:
            raise ImportError('this loader can only load the ansible_collections toplevel package, not {0}'.format(self._fullname))


# Implements Ansible's custom namespace package support.
# The ansible_collections package and one level down (collections namespaces) are Python namespace packages
# that search across all configured collection roots. The collection package (two levels down) is the first one found
# on the configured collection root path, and Python namespace package aggregation is not allowed at or below
# the collection. Implements implicit package (package dir) support for both Py2/3. Package init code is ignored
# by this loader.
class _AnsibleCollectionNSPkgLoader(_AnsibleCollectionPkgLoaderBase):
    def _validate_args(self):
        super(_AnsibleCollectionNSPkgLoader, self)._validate_args()
        if len(self._split_name) != 2:
            raise ImportError('this loader can only load collections namespace packages, not {0}'.format(self._fullname))

    def _validate_final(self):
        # special-case the `ansible` namespace, since `ansible.builtin` is magical
        if not self._subpackage_search_paths and self._package_to_load != 'ansible':
            raise ImportError('no {0} found in {1}'.format(self._package_to_load, self._candidate_paths))


# handles locating the actual collection package and associated metadata
class _AnsibleCollectionPkgLoader(_AnsibleCollectionPkgLoaderBase):
    def _validate_args(self):
        super(_AnsibleCollectionPkgLoader, self)._validate_args()
        if len(self._split_name) != 3:
            raise ImportError('this loader can only load collection packages, not {0}'.format(self._fullname))

    def _validate_final(self):
        if self._split_name[1:3] == ['ansible', 'builtin']:
            # we don't want to allow this one to have on-disk search capability
            self._subpackage_search_paths = []
        elif not self._subpackage_search_paths:
            raise ImportError('no {0} found in {1}'.format(self._package_to_load, self._candidate_paths))
        else:
            # only search within the first collection we found
            self._subpackage_search_paths = [self._subpackage_search_paths[0]]

    def _load_module(self, module):
        if not _meta_yml_to_dict:
            raise ValueError('ansible.utils.collection_loader._meta_yml_to_dict is not set')

        module._collection_meta = {}
        # TODO: load collection metadata, cache in __loader__ state

        collection_name = '.'.join(self._split_name[1:3])

        if collection_name == 'ansible.builtin':
            # ansible.builtin is a synthetic collection, get its routing config from the Ansible distro
            ansible_pkg_path = os.path.dirname(import_module('ansible').__file__)
            metadata_path = os.path.join(ansible_pkg_path, 'config/ansible_builtin_runtime.yml')
            with open(_to_bytes(metadata_path), 'rb') as fd:
                raw_routing = fd.read()
        else:
            b_routing_meta_path = _to_bytes(os.path.join(module.__path__[0], 'meta/runtime.yml'))
            if os.path.isfile(b_routing_meta_path):
                with open(b_routing_meta_path, 'rb') as fd:
                    raw_routing = fd.read()
            else:
                raw_routing = ''
        try:
            if raw_routing:
                routing_dict = _meta_yml_to_dict(raw_routing, (collection_name, 'runtime.yml'))
                module._collection_meta = self._canonicalize_meta(routing_dict)
        except Exception as ex:
            raise ValueError(f'error parsing collection metadata: {ex}')

        AnsibleCollectionConfig.on_collection_load.fire(collection_name=collection_name, collection_path=os.path.dirname(module.__file__))

        return module

    def exec_module(self, module):
        super(_AnsibleCollectionPkgLoader, self).exec_module(module)
        self._load_module(module)

    def create_module(self, spec):
        return None

    def load_module(self, fullname):
        module = super(_AnsibleCollectionPkgLoader, self).load_module(fullname)
        return self._load_module(module)

    def _canonicalize_meta(self, meta_dict):
        # TODO: rewrite import keys and all redirect targets that start with .. (current namespace) and . (current collection)
        # OR we could do it all on the fly?
        # if not meta_dict:
        #     return {}
        #
        # ns_name = '.'.join(self._split_name[0:2])
        # collection_name = '.'.join(self._split_name[0:3])
        #
        # #
        # for routing_type, routing_type_dict in iteritems(meta_dict.get('plugin_routing', {})):
        #     for plugin_key, plugin_dict in iteritems(routing_type_dict):
        #         redirect = plugin_dict.get('redirect', '')
        #         if redirect.startswith('..'):
        #             redirect =  redirect[2:]

        return meta_dict


# loads everything under a collection, including handling redirections defined by the collection
class _AnsibleCollectionLoader(_AnsibleCollectionPkgLoaderBase):
    # HACK: stash this in a better place
    _redirected_package_map = {}  # type: dict[str, str]
    _allows_package_code = True

    def _validate_args(self):
        super(_AnsibleCollectionLoader, self)._validate_args()
        if len(self._split_name) < 4:
            raise ValueError('this loader is only for sub-collection modules/packages, not {0}'.format(self._fullname))

    def _get_candidate_paths(self, path_list):
        if len(path_list) != 1 and self._split_name[1:3] != ['ansible', 'builtin']:
            raise ValueError('this loader requires exactly one path to search')

        return path_list

    def _get_subpackage_search_paths(self, candidate_paths):
        collection_name = '.'.join(self._split_name[1:3])
        collection_meta = _get_collection_metadata(collection_name)

        # check for explicit redirection, as well as ancestor package-level redirection (only load the actual code once!)
        redirect = None
        explicit_redirect = False

        routing_entry = _nested_dict_get(collection_meta, ['import_redirection', self._fullname])
        if routing_entry:
            redirect = routing_entry.get('redirect')

        if redirect:
            explicit_redirect = True
        else:
            redirect = _get_ancestor_redirect(self._redirected_package_map, self._fullname)

        # NB: package level redirection requires hooking all future imports beneath the redirected source package
        # in order to ensure sanity on future relative imports. We always import everything under its "real" name,
        # then add a sys.modules entry with the redirected name using the same module instance. If we naively imported
        # the source for each redirection, most submodules would import OK, but we'd have N runtime copies of the module
        # (one for each name), and relative imports that ascend above the redirected package would break (since they'd
        # see the redirected ancestor package contents instead of the package where they actually live).
        if redirect:
            # FIXME: wrap this so we can be explicit about a failed redirection
            self._redirect_module = import_module(redirect)
            if explicit_redirect and hasattr(self._redirect_module, '__path__') and self._redirect_module.__path__:
                # if the import target looks like a package, store its name so we can rewrite future descendent loads
                self._redirected_package_map[self._fullname] = redirect

            # if we redirected, don't do any further custom package logic
            return None

        # we're not doing a redirect- try to find what we need to actually load a module/package

        # this will raise ImportError if we can't find the requested module/package at all
        if not candidate_paths:
            # noplace to look, just ImportError
            raise ImportError('package has no paths')

        found_path, has_code, package_path = self._module_file_from_path(self._package_to_load, candidate_paths[0])

        # still here? we found something to load...
        if has_code:
            self._source_code_path = found_path

        if package_path:
            return [package_path]  # always needs to be a list

        return None


# This loader only answers for intercepted Ansible Python modules. Normal imports will fail here and be picked up later
# by our path_hook importer (which proxies the built-in import mechanisms, allowing normal caching etc to occur)
class _AnsibleInternalRedirectLoader:
    def __init__(self, fullname, path_list):
        self._redirect = None

        split_name = fullname.split('.')
        toplevel_pkg = split_name[0]
        module_to_load = split_name[-1]

        if toplevel_pkg != 'ansible':
            raise ImportError('not interested')

        builtin_meta = _get_collection_metadata('ansible.builtin')

        routing_entry = _nested_dict_get(builtin_meta, ['import_redirection', fullname])
        if routing_entry:
            self._redirect = routing_entry.get('redirect')

        if not self._redirect:
            raise ImportError('not redirected, go ask path_hook')

    def get_resource_reader(self, fullname):
        return _AnsibleTraversableResources(fullname, self)

    def exec_module(self, module):
        # should never see this
        if not self._redirect:
            raise ValueError('no redirect found for {0}'.format(module.__spec__.name))

        # Replace the module with the redirect
        sys.modules[module.__spec__.name] = import_module(self._redirect)

    def create_module(self, spec):
        return None

    def load_module(self, fullname):
        # since we're delegating to other loaders, this should only be called for internal redirects where we answered
        # find_module with this loader, in which case we'll just directly import the redirection target, insert it into
        # sys.modules under the name it was requested by, and return the original module.

        # should never see this
        if not self._redirect:
            raise ValueError('no redirect found for {0}'.format(fullname))

        # FIXME: smuggle redirection context, provide warning/error that we tried and failed to redirect
        mod = import_module(self._redirect)
        sys.modules[fullname] = mod
        return mod


class AnsibleCollectionRef:
    # FUTURE: introspect plugin loaders to get these dynamically?
    VALID_REF_TYPES = frozenset(_to_text(r) for r in ['action', 'become', 'cache', 'callback', 'cliconf', 'connection',
                                                      'doc_fragments', 'filter', 'httpapi', 'inventory', 'lookup',
                                                      'module_utils', 'modules', 'netconf', 'role', 'shell', 'strategy',
                                                      'terminal', 'test', 'vars', 'playbook'])

    # FIXME: tighten this up to match Python identifier reqs, etc
    VALID_SUBDIRS_RE = re.compile(_to_text(r'^\w+(\.\w+)*$'))
    VALID_FQCR_RE = re.compile(_to_text(r'^\w+(\.\w+){2,}$'))  # can have 0-N included subdirs as well

    def __init__(self, collection_name, subdirs, resource, ref_type):
        """
        Create an AnsibleCollectionRef from components
        :param collection_name: a collection name of the form 'namespace.collectionname'
        :param subdirs: optional subdir segments to be appended below the plugin type (eg, 'subdir1.subdir2')
        :param resource: the name of the resource being references (eg, 'mymodule', 'someaction', 'a_role')
        :param ref_type: the type of the reference, eg 'module', 'role', 'doc_fragment'
        """
        collection_name = _to_text(collection_name, strict=True)
        if subdirs is not None:
            subdirs = _to_text(subdirs, strict=True)
        resource = _to_text(resource, strict=True)
        ref_type = _to_text(ref_type, strict=True)

        if not self.is_valid_collection_name(collection_name):
            raise ValueError('invalid collection name (must be of the form namespace.collection): {0}'.format(_to_text(collection_name)))

        if ref_type not in self.VALID_REF_TYPES:
            raise ValueError('invalid collection ref_type: {0}'.format(ref_type))

        self.collection = collection_name
        if subdirs:
            if not re.match(self.VALID_SUBDIRS_RE, subdirs):
                raise ValueError('invalid subdirs entry: {0} (must be empty/None or of the form subdir1.subdir2)'.format(_to_text(subdirs)))
            self.subdirs = subdirs
        else:
            self.subdirs = u''

        self.resource = resource
        self.ref_type = ref_type

        package_components = [u'ansible_collections', self.collection]
        fqcr_components = [self.collection]

        self.n_python_collection_package_name = _to_text('.'.join(package_components))

        if self.ref_type == u'role':
            package_components.append(u'roles')
        elif self.ref_type == u'playbook':
            package_components.append(u'playbooks')
        else:
            # we assume it's a plugin
            package_components += [u'plugins', self.ref_type]

        if self.subdirs:
            package_components.append(self.subdirs)
            fqcr_components.append(self.subdirs)

        if self.ref_type in (u'role', u'playbook'):
            # playbooks and roles are their own resource
            package_components.append(self.resource)

        fqcr_components.append(self.resource)

        self.n_python_package_name = _to_text('.'.join(package_components))
        self._fqcr = u'.'.join(fqcr_components)

    def __repr__(self):
        return 'AnsibleCollectionRef(collection={0!r}, subdirs={1!r}, resource={2!r})'.format(self.collection, self.subdirs, self.resource)

    @property
    def fqcr(self):
        return self._fqcr

    @staticmethod
    def from_fqcr(ref, ref_type):
        """
        Parse a string as a fully-qualified collection reference, raises ValueError if invalid
        :param ref: collection reference to parse (a valid ref is of the form 'ns.coll.resource' or 'ns.coll.subdir1.subdir2.resource')
        :param ref_type: the type of the reference, eg 'module', 'role', 'doc_fragment'
        :return: a populated AnsibleCollectionRef object
        """
        # assuming the fq_name is of the form (ns).(coll).(optional_subdir_N).(resource_name),
        # we split the resource name off the right, split ns and coll off the left, and we're left with any optional
        # subdirs that need to be added back below the plugin-specific subdir we'll add. So:
        # ns.coll.resource -> ansible_collections.ns.coll.plugins.(plugintype).resource
        # ns.coll.subdir1.resource -> ansible_collections.ns.coll.plugins.subdir1.(plugintype).resource
        # ns.coll.rolename -> ansible_collections.ns.coll.roles.rolename
        if not AnsibleCollectionRef.is_valid_fqcr(ref):
            raise ValueError('{0} is not a valid collection reference'.format(_to_text(ref)))

        ref = _to_text(ref, strict=True)
        ref_type = _to_text(ref_type, strict=True)
        ext = ''

        if ref_type == u'playbook' and ref.endswith(PB_EXTENSIONS):
            resource_splitname = ref.rsplit(u'.', 2)
            package_remnant = resource_splitname[0]
            resource = resource_splitname[1]
            ext = '.' + resource_splitname[2]
        else:
            resource_splitname = ref.rsplit(u'.', 1)
            package_remnant = resource_splitname[0]
            resource = resource_splitname[1]

        # split the left two components of the collection package name off, anything remaining is plugin-type
        # specific subdirs to be added back on below the plugin type
        package_splitname = package_remnant.split(u'.', 2)
        if len(package_splitname) == 3:
            subdirs = package_splitname[2]
        else:
            subdirs = u''

        collection_name = u'.'.join(package_splitname[0:2])

        return AnsibleCollectionRef(collection_name, subdirs, resource + ext, ref_type)

    @staticmethod
    def try_parse_fqcr(ref, ref_type):
        """
        Attempt to parse a string as a fully-qualified collection reference, returning None on failure (instead of raising an error)
        :param ref: collection reference to parse (a valid ref is of the form 'ns.coll.resource' or 'ns.coll.subdir1.subdir2.resource')
        :param ref_type: the type of the reference, eg 'module', 'role', 'doc_fragment'
        :return: a populated AnsibleCollectionRef object on successful parsing, else None
        """
        try:
            return AnsibleCollectionRef.from_fqcr(ref, ref_type)
        except ValueError:
            pass

    @staticmethod
    def legacy_plugin_dir_to_plugin_type(legacy_plugin_dir_name):
        """
        Utility method to convert from a PluginLoader dir name to a plugin ref_type
        :param legacy_plugin_dir_name: PluginLoader dir name (eg, 'action_plugins', 'library')
        :return: the corresponding plugin ref_type (eg, 'action', 'role')
        """
        if legacy_plugin_dir_name is None:
            plugin_type = None
        else:
            legacy_plugin_dir_name = _to_text(legacy_plugin_dir_name)

            plugin_type = legacy_plugin_dir_name.removesuffix(u'_plugins')

        if plugin_type == u'library':
            plugin_type = u'modules'

        if plugin_type not in AnsibleCollectionRef.VALID_REF_TYPES:
            raise ValueError(f'{legacy_plugin_dir_name!r} cannot be mapped to a valid collection ref type')

        return plugin_type

    @staticmethod
    def is_valid_fqcr(ref, ref_type=None):
        """
        Validates if is string is a well-formed fully-qualified collection reference (does not look up the collection itself)
        :param ref: candidate collection reference to validate (a valid ref is of the form 'ns.coll.resource' or 'ns.coll.subdir1.subdir2.resource')
        :param ref_type: optional reference type to enable deeper validation, eg 'module', 'role', 'doc_fragment'
        :return: True if the collection ref passed is well-formed, False otherwise
        """

        ref = _to_text(ref)

        if not ref_type:
            return bool(re.match(AnsibleCollectionRef.VALID_FQCR_RE, ref))

        return bool(AnsibleCollectionRef.try_parse_fqcr(ref, ref_type))

    @staticmethod
    def is_valid_collection_name(collection_name):
        """
        Validates if the given string is a well-formed collection name (does not look up the collection itself)
        :param collection_name: candidate collection name to validate (a valid name is of the form 'ns.collname')
        :return: True if the collection name passed is well-formed, False otherwise
        """

        collection_name = _to_text(collection_name)

        if collection_name.count(u'.') != 1:
            return False

        return all(
            # NOTE: keywords and identifiers are different in different Pythons
            not iskeyword(ns_or_name) and ns_or_name.isidentifier()
            for ns_or_name in collection_name.split(u'.')
        )


def _get_collection_path(collection_name):
    collection_name = _to_text(collection_name)
    if not collection_name or not isinstance(collection_name, str) or len(collection_name.split('.')) != 2:
        raise ValueError('collection_name must be a non-empty string of the form namespace.collection')
    try:
        collection_pkg = import_module('ansible_collections.' + collection_name)
    except ImportError:
        raise ValueError('unable to locate collection {0}'.format(collection_name))

    return _to_text(os.path.dirname(_to_bytes(collection_pkg.__file__)))


def _get_collection_playbook_path(playbook):

    acr = AnsibleCollectionRef.try_parse_fqcr(playbook, u'playbook')
    if acr:
        try:
            # get_collection_path
            pkg = import_module(acr.n_python_collection_package_name)
        except (OSError, ModuleNotFoundError) as ex:
            # leaving ex as debug target, even though not used in normal code
            pkg = None

        if pkg:
            cpath = os.path.join(sys.modules[acr.n_python_collection_package_name].__file__.replace('__synthetic__', 'playbooks'))

            if acr.subdirs:
                paths = [_to_text(x) for x in acr.subdirs.split(u'.')]
                paths.insert(0, cpath)
                cpath = os.path.join(*paths)

            path = os.path.join(cpath, _to_text(acr.resource))
            if os.path.exists(_to_bytes(path)):
                return acr.resource, path, acr.collection
            elif not acr.resource.endswith(PB_EXTENSIONS):
                for ext in PB_EXTENSIONS:
                    path = os.path.join(cpath, _to_text(acr.resource + ext))
                    if os.path.exists(_to_bytes(path)):
                        return acr.resource, path, acr.collection
    return None


def _get_collection_role_path(role_name, collection_list=None):
    return _get_collection_resource_path(role_name, u'role', collection_list)


def _get_collection_resource_path(name, ref_type, collection_list=None):

    if ref_type == u'playbook':
        # they are handled a bit diff due to 'extension variance' and no collection_list
        return _get_collection_playbook_path(name)

    acr = AnsibleCollectionRef.try_parse_fqcr(name, ref_type)
    if acr:
        # looks like a valid qualified collection ref; skip the collection_list
        collection_list = [acr.collection]
        subdirs = acr.subdirs
        resource = acr.resource
    elif not collection_list:
        return None  # not a FQ and no collection search list spec'd, nothing to do
    else:
        resource = name  # treat as unqualified, loop through the collection search list to try and resolve
        subdirs = ''

    for collection_name in collection_list:
        try:
            acr = AnsibleCollectionRef(collection_name=collection_name, subdirs=subdirs, resource=resource, ref_type=ref_type)
            # FIXME: error handling/logging; need to catch any import failures and move along
            pkg = import_module(acr.n_python_package_name)

            if pkg is not None:
                # the package is now loaded, get the collection's package and ask where it lives
                path = os.path.dirname(_to_bytes(sys.modules[acr.n_python_package_name].__file__))
                return resource, _to_text(path), collection_name

        except (OSError, ModuleNotFoundError) as ex:
            continue
        except Exception as ex:
            # FIXME: pick out typical import errors first, then error logging
            continue

    return None


def _get_collection_name_from_path(path):
    """
    Return the containing collection name for a given path, or None if the path is not below a configured collection, or
    the collection cannot be loaded (eg, the collection is masked by another of the same name higher in the configured
    collection roots).
    :param path: path to evaluate for collection containment
    :return: collection name or None
    """

    # ensure we compare full paths since pkg path will be abspath
    path = _to_text(os.path.abspath(_to_bytes(path)))

    path_parts = path.split('/')
    if path_parts.count('ansible_collections') != 1:
        return None

    ac_pos = path_parts.index('ansible_collections')

    # make sure it's followed by at least a namespace and collection name
    if len(path_parts) < ac_pos + 3:
        return None

    candidate_collection_name = '.'.join(path_parts[ac_pos + 1:ac_pos + 3])

    try:
        # we've got a name for it, now see if the path prefix matches what the loader sees
        imported_pkg_path = _to_text(os.path.dirname(_to_bytes(import_module('ansible_collections.' + candidate_collection_name).__file__)))
    except ImportError:
        return None

    # reassemble the original path prefix up the collection name, and it should match what we just imported. If not
    # this is probably a collection root that's not configured.

    original_path_prefix = os.path.join('/', *path_parts[0:ac_pos + 3])

    imported_pkg_path = _to_text(os.path.abspath(_to_bytes(imported_pkg_path)))
    if original_path_prefix != imported_pkg_path:
        return None

    return candidate_collection_name


def _get_import_redirect(collection_meta_dict, fullname):
    if not collection_meta_dict:
        return None

    return _nested_dict_get(collection_meta_dict, ['import_redirection', fullname, 'redirect'])


def _get_ancestor_redirect(redirected_package_map, fullname):
    # walk the requested module's ancestor packages to see if any have been previously redirected
    cur_pkg = fullname
    while cur_pkg:
        cur_pkg = cur_pkg.rpartition('.')[0]
        ancestor_redirect = redirected_package_map.get(cur_pkg)
        if ancestor_redirect:
            # rewrite the prefix on fullname so we import the target first, then alias it
            redirect = ancestor_redirect + fullname[len(cur_pkg):]
            return redirect
    return None


def _nested_dict_get(root_dict, key_list):
    cur_value = root_dict
    for key in key_list:
        cur_value = cur_value.get(key)
        if not cur_value:
            return None

    return cur_value


def _iter_modules_impl(paths, prefix=''):
    # NB: this currently only iterates what's on disk- redirected modules are not considered
    if not prefix:
        prefix = ''
    else:
        prefix = _to_text(prefix)
    # yield (module_loader, name, ispkg) for each module/pkg under path
    # TODO: implement ignore/silent catch for unreadable?
    for b_path in map(_to_bytes, paths):
        if not os.path.isdir(b_path):
            continue
        for b_basename in sorted(os.listdir(b_path)):
            b_candidate_module_path = os.path.join(b_path, b_basename)
            if os.path.isdir(b_candidate_module_path):
                # exclude things that obviously aren't Python package dirs
                # FIXME: this dir is adjustable in py3.8+, check for it
                if b'.' in b_basename or b_basename == b'__pycache__':
                    continue

                # TODO: proper string handling?
                yield prefix + _to_text(b_basename), True
            else:
                # FIXME: match builtin ordering for package/dir/file, support compiled?
                if b_basename.endswith(b'.py') and b_basename != b'__init__.py':
                    yield prefix + _to_text(os.path.splitext(b_basename)[0]), False


def _get_collection_metadata(collection_name):
    collection_name = _to_text(collection_name)
    if not collection_name or not isinstance(collection_name, str) or len(collection_name.split('.')) != 2:
        raise ValueError('collection_name must be a non-empty string of the form namespace.collection')

    try:
        collection_pkg = import_module('ansible_collections.' + collection_name)
    except ImportError as ex:
        raise ValueError('unable to locate collection {0}'.format(collection_name)) from ex

    _collection_meta = getattr(collection_pkg, '_collection_meta', None)

    if _collection_meta is None:
        raise ValueError('collection metadata was not loaded for collection {0}'.format(collection_name))

    return _collection_meta



from __future__ import annotations

import typing as t

import pytest

from ansible._internal import _display_utils
from ansible._internal._datatag._tags import Origin
from ansible.module_utils._internal._datatag import AnsibleTagHelper
from ansible.parsing.vault import EncryptedString
from ansible.parsing.yaml.objects import _AnsibleMapping, _AnsibleUnicode, _AnsibleSequence
from ansible.parsing.yaml import objects


@pytest.fixture(autouse=True, scope='function')
def suppress_warnings() -> t.Generator[None]:
    with _display_utils.DeferredWarningContext(variables={}):
        yield


def test_ansible_mapping() -> None:
    from ansible.parsing.yaml.objects import AnsibleMapping

    value = dict(a=1)
    result = AnsibleMapping(value)

    assert type(result) is type(value)  # pylint: disable=unidiomatic-typecheck
    assert result == value


def test_tagged_ansible_mapping() -> None:
    from ansible.parsing.yaml.objects import AnsibleMapping

    value = Origin(description='test').tag(dict(a=1))
    result = AnsibleMapping(value)

    assert type(result) is type(value)  # pylint: disable=unidiomatic-typecheck
    assert result == value
    assert AnsibleTagHelper.tags(result) == AnsibleTagHelper.tags(value)


def test_ansible_unicode() -> None:
    from ansible.parsing.yaml.objects import AnsibleUnicode

    value = 'hello'
    result = AnsibleUnicode(value)

    assert type(result) is type(value)  # pylint: disable=unidiomatic-typecheck
    assert result == value


def test_tagged_ansible_unicode() -> None:
    from ansible.parsing.yaml.objects import AnsibleUnicode

    value = Origin(description='test').tag('hello')
    result = AnsibleUnicode(value)

    assert type(result) is type(value)  # pylint: disable=unidiomatic-typecheck
    assert result == value
    assert AnsibleTagHelper.tags(result) == AnsibleTagHelper.tags(value)


def test_ansible_sequence() -> None:
    from ansible.parsing.yaml.objects import AnsibleSequence

    value = [1, 2, 3]
    result = AnsibleSequence(value)

    assert type(result) is type(value)  # pylint: disable=unidiomatic-typecheck
    assert result == value


def test_tagged_ansible_sequence() -> None:
    from ansible.parsing.yaml.objects import AnsibleSequence

    value = Origin(description='test').tag([1, 2, 3])
    result = AnsibleSequence(value)

    assert type(result) is type(value)  # pylint: disable=unidiomatic-typecheck
    assert result == value
    assert AnsibleTagHelper.tags(result) == AnsibleTagHelper.tags(value)


def test_ansible_vault_encrypted_unicode() -> None:
    from ansible.parsing.yaml.objects import AnsibleVaultEncryptedUnicode

    value = 'ciphertext'
    result = AnsibleVaultEncryptedUnicode(value)

    assert type(result) is EncryptedString  # pylint: disable=unidiomatic-typecheck
    assert result._ciphertext == value


def test_tagged_ansible_vault_encrypted_unicode() -> None:
    from ansible.parsing.yaml.objects import AnsibleVaultEncryptedUnicode

    value = Origin(description='test').tag('ciphertext')
    result = AnsibleVaultEncryptedUnicode(value)

    assert type(result) is EncryptedString  # pylint: disable=unidiomatic-typecheck
    assert result._ciphertext == value
    assert AnsibleTagHelper.tags(result) == AnsibleTagHelper.tags(value)


def test_invalid_attribute() -> None:
    with pytest.raises(ImportError, match="cannot import name 'bogus' from 'ansible.parsing.yaml.objects'"):
        from ansible.parsing.yaml.objects import bogus

    with pytest.raises(AttributeError, match="module 'ansible.parsing.yaml.objects' has no attribute 'bogus'"):
        assert objects.bogus


def test_non_ansible_attribute() -> None:
    with pytest.raises(ImportError, match="cannot import name 't' from 'ansible.parsing.yaml.objects'"):
        from ansible.parsing.yaml.objects import t

    with pytest.raises(AttributeError, match="module 'ansible.parsing.yaml.objects' has no attribute 't'"):
        assert objects.t


@pytest.mark.parametrize("target_type,args,kwargs,expected", (
    (_AnsibleMapping, (), {}, {}),
    (_AnsibleMapping, (dict(a=1),), {}, dict(a=1)),
    (_AnsibleMapping, (dict(a=1),), dict(b=2), dict(a=1, b=2)),
    (_AnsibleUnicode, (), {}, ''),
    (_AnsibleUnicode, ('Hello',), {}, 'Hello'),
    (_AnsibleUnicode, (), dict(object='Hello'), 'Hello'),
    (_AnsibleUnicode, (b'Hello',), {}, str(b'Hello')),
    (_AnsibleUnicode, (b'Hello',), dict(encoding='utf-8', errors='strict'), 'Hello'),
    (_AnsibleSequence, (), {}, []),
    (_AnsibleSequence, ([1, 2],), {}, [1, 2]),
))
def test_objects(target_type: type, args: tuple, kwargs: dict, expected: object) -> None:
    """Verify legacy objects support the same constructor args as their base types."""
    result = target_type(*args, **kwargs)

    assert isinstance(result, type(expected))
    assert result == expected



from __future__ import annotations

import io
import typing as _t

from .._wrapt import ObjectProxy
from ...module_utils._internal import _datatag


class TaggedStreamWrapper(ObjectProxy):
    """
    Janky proxy around IOBase to allow streams to carry tags and support basic interrogation by the tagging API.
    Most tagging operations will have undefined behavior for this type.
    """

    _self__ansible_tags_mapping: _datatag._AnsibleTagsMapping

    def __init__(self, stream: io.IOBase, tags: _datatag.AnsibleDatatagBase | _t.Iterable[_datatag.AnsibleDatatagBase]) -> None:
        super().__init__(stream)

        tag_list: list[_datatag.AnsibleDatatagBase]

        # noinspection PyProtectedMember
        if type(tags) in _datatag._known_tag_types:
            tag_list = [tags]  # type: ignore[list-item]
        else:
            tag_list = list(tags)  # type: ignore[arg-type]

        self._self__ansible_tags_mapping = _datatag._AnsibleTagsMapping((type(tag), tag) for tag in tag_list)

    @property
    def _ansible_tags_mapping(self) -> _datatag._AnsibleTagsMapping:
        return self._self__ansible_tags_mapping



from __future__ import annotations

import collections.abc as _c
import dataclasses
import typing as t

from ansible._internal._errors import _error_utils
from ansible.errors import AnsibleRuntimeError
from ansible.module_utils._internal import _messages


class AnsibleCapturedError(AnsibleRuntimeError):
    """An exception representing error detail captured in another context where the error detail must be serialized to be preserved."""

    context: t.ClassVar[str]

    def __init__(
        self,
        *,
        obj: t.Any = None,
        event: _messages.Event,
    ) -> None:
        super().__init__(
            obj=obj,
        )

        self._event = event


class AnsibleResultCapturedError(AnsibleCapturedError, _error_utils.ContributesToTaskResult):
    """
    An exception representing error detail captured in a foreign context where an action/module result dictionary is involved.

    This exception provides a result dictionary via the ContributesToTaskResult mixin.
    """

    def __init__(self, event: _messages.Event, result: dict[str, t.Any]) -> None:
        super().__init__(event=event)

        self._result = result

    @property
    def result_contribution(self) -> _c.Mapping[str, object]:
        return self._result

    @classmethod
    def maybe_raise_on_result(cls, result: dict[str, t.Any]) -> None:
        """Normalize the result and raise an exception if the result indicated failure."""
        if error_summary := cls.normalize_result_exception(result):
            raise error_summary.error_type(error_summary.event, result)

    @classmethod
    def normalize_result_exception(cls, result: dict[str, t.Any]) -> CapturedErrorSummary | None:
        """
        Normalize the result `exception`, if any, to be a `CapturedErrorSummary` instance.
        If a new `CapturedErrorSummary` was created, the `error_type` will be `cls`.
        The `exception` key will be removed if falsey.
        A `CapturedErrorSummary` instance will be returned if `failed` is truthy.
        """
        if type(cls) is AnsibleResultCapturedError:  # pylint: disable=unidiomatic-typecheck
            raise TypeError('The normalize_result_exception method cannot be called on the AnsibleCapturedError base type, use a derived type.')

        if not isinstance(result, dict):
            raise TypeError(f'Malformed result. Received {type(result)} instead of {dict}.')

        failed = result.get('failed')  # DTFIX-FUTURE: warn if failed is present and not a bool, or exception is present without failed being True
        exception = result.pop('exception', None)

        if not failed and not exception:
            return None

        if isinstance(exception, CapturedErrorSummary):
            error_summary = exception
        elif isinstance(exception, _messages.ErrorSummary):
            error_summary = CapturedErrorSummary(
                event=exception.event,
                error_type=cls,
            )
        else:
            # translate non-ErrorDetail errors
            error_summary = CapturedErrorSummary(
                event=_messages.Event(
                    msg=str(result.get('msg', 'Unknown error.')),
                    formatted_traceback=cls._normalize_traceback(exception),
                ),
                error_type=cls,
            )

        result.update(exception=error_summary)

        return error_summary if failed else None  # even though error detail was normalized, only return it if the result indicated failure

    @classmethod
    def _normalize_traceback(cls, value: object | None) -> str | None:
        """Normalize the provided traceback value, returning None if it is falsey."""
        if not value:
            return None

        value = str(value).rstrip()

        if not value:
            return None

        return value + '\n'


class AnsibleActionCapturedError(AnsibleResultCapturedError):
    """An exception representing error detail sourced directly by an action in its result dictionary."""

    _default_message = 'Action failed.'
    context = 'action'


class AnsibleModuleCapturedError(AnsibleResultCapturedError):
    """An exception representing error detail captured in a module context and returned from an action's result dictionary."""

    _default_message = 'Module failed.'
    context = 'target'


@dataclasses.dataclass(**_messages._dataclass_kwargs)
class CapturedErrorSummary(_messages.ErrorSummary):
    error_type: type[AnsibleResultCapturedError] | None = None



# (c) Ansible Project

from __future__ import annotations

from urllib.parse import urlparse


def is_uri(value, schemes=None):
    """ Will verify that the string passed is a valid 'URI', if given a list of valid schemes it will match those """
    try:
        x = urlparse(value)
        isit = all([x.scheme is not None, x.path is not None, not schemes or x.scheme in schemes])
    except Exception as e:
        isit = False
    return isit


def is_url(value, schemes=None):
    """ Will verify that the string passed is a valid 'URL' """

    isit = is_uri(value, schemes)
    if isit:
        x = urlparse(value)
        isit = bool(x.netloc or x.scheme == 'file')
    return isit


def is_urn(value):
    return is_uri(value, ['urn'])


class TestModule(object):
    """ Ansible URI jinja2 test """

    def tests(self):
        return {
            # file testing
            'uri': is_uri,
            'url': is_url,
            'urn': is_urn,
        }



from __future__ import annotations

import dataclasses

from ansible.module_utils._internal import _ambient_context, _messages
from . import _event_formatting


class DeferredWarningContext(_ambient_context.AmbientContextBase):
    """
    Calls to `Display.warning()` and `Display.deprecated()` within this context will cause the resulting warnings to be captured and not displayed.
    The intended use is for task-initiated warnings to be recorded with the task result, which makes them visible to registered results, callbacks, etc.
    The active display callback is responsible for communicating any warnings to the user.
    """

    # DTFIX-FUTURE: once we start implementing nested scoped contexts for our own bookkeeping, this should be an interface facade that forwards to the nearest
    #               context that actually implements the warnings collection capability

    def __init__(self, *, variables: dict[str, object]) -> None:
        self._variables = variables  # DTFIX-FUTURE: move this to an AmbientContext-derived TaskContext (once it exists)
        self._deprecation_warnings: list[_messages.DeprecationSummary] = []
        self._warnings: list[_messages.WarningSummary] = []
        self._seen: set[_messages.WarningSummary] = set()

    def capture(self, warning: _messages.WarningSummary) -> None:
        """Add the warning/deprecation to the context if it has not already been seen by this context."""
        if warning in self._seen:
            return

        self._seen.add(warning)

        if isinstance(warning, _messages.DeprecationSummary):
            self._deprecation_warnings.append(warning)
        else:
            self._warnings.append(warning)

    def get_warnings(self) -> list[_messages.WarningSummary]:
        """Return a list of the captured non-deprecation warnings."""
        # DTFIX-FUTURE: return a read-only list proxy instead
        return self._warnings

    def get_deprecation_warnings(self) -> list[_messages.DeprecationSummary]:
        """Return a list of the captured deprecation warnings."""
        # DTFIX-FUTURE: return a read-only list proxy instead
        return self._deprecation_warnings


def format_message(summary: _messages.SummaryBase, include_traceback: bool) -> str:
    if isinstance(summary, _messages.DeprecationSummary):
        deprecation_message = get_deprecation_message_with_plugin_info(
            msg=summary.event.msg,
            version=summary.version,
            date=summary.date,
            deprecator=summary.deprecator,
        )

        event = dataclasses.replace(summary.event, msg=deprecation_message)
    else:
        event = summary.event

    return _event_formatting.format_event(event, include_traceback)


def get_deprecation_message_with_plugin_info(
    *,
    msg: str,
    version: str | None,
    removed: bool = False,
    date: str | None,
    deprecator: _messages.PluginInfo | None,
) -> str:
    """Internal use only. Return a deprecation message and help text for display."""
    # DTFIX-FUTURE: the logic for omitting date/version doesn't apply to the payload, so it shows up in vars in some cases when it should not

    if removed:
        removal_fragment = 'This feature was removed'
    else:
        removal_fragment = 'This feature will be removed'

    if not deprecator or not deprecator.type:
        # indeterminate has no resolved_name or type
        # collections have a resolved_name but no type
        collection = deprecator.resolved_name if deprecator else None
        plugin_fragment = ''
    elif deprecator.resolved_name == 'ansible.builtin':
        # core deprecations from base classes (the API) have no plugin name, only 'ansible.builtin'
        plugin_type_name = str(deprecator.type) if deprecator.type is _messages.PluginType.MODULE else f'{deprecator.type} plugin'

        collection = deprecator.resolved_name
        plugin_fragment = f'the {plugin_type_name} API'
    else:
        parts = deprecator.resolved_name.split('.')
        plugin_name = parts[-1]
        plugin_type_name = str(deprecator.type) if deprecator.type is _messages.PluginType.MODULE else f'{deprecator.type} plugin'

        collection = '.'.join(parts[:2]) if len(parts) > 2 else None
        plugin_fragment = f'{plugin_type_name} {plugin_name!r}'

    if collection and plugin_fragment:
        plugin_fragment += ' in'

    if collection == 'ansible.builtin':
        collection_fragment = 'ansible-core'
    elif collection:
        collection_fragment = f'collection {collection!r}'
    else:
        collection_fragment = ''

    if not collection:
        when_fragment = 'in the future' if not removed else ''
    elif date:
        when_fragment = f'in a release after {date}'
    elif version:
        when_fragment = f'version {version}'
    else:
        when_fragment = 'in a future release' if not removed else ''

    if plugin_fragment or collection_fragment:
        from_fragment = 'from'
    else:
        from_fragment = ''

    deprecation_msg = ' '.join(f for f in [removal_fragment, from_fragment, plugin_fragment, collection_fragment, when_fragment] if f) + '.'

    return join_sentences(msg, deprecation_msg)


def join_sentences(first: str | None, second: str | None) -> str:
    """Join two sentences together."""
    first = (first or '').strip()
    second = (second or '').strip()

    if first and first[-1] not in ('!', '?', '.'):
        first += '.'

    if second and second[-1] not in ('!', '?', '.'):
        second += '.'

    if first and not second:
        return first

    if not first and second:
        return second

    return ' '.join((first, second))



from __future__ import annotations


def override_formerly_core_masked_filter(*args, **kwargs):
    return 'hello from overridden formerly_core_masked_filter'


class FilterModule(object):
    def filters(self):
        return {
            'formerly_core_masked_filter': override_formerly_core_masked_filter
        }



# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = """
name: world
short_description: World lookup
description: A world lookup.
author:
  - Ansible Core Team
"""

EXAMPLES = """
- debug:
    msg: "{{ lookup('ns.col.world') }}"
"""

RETURN = """ # """

from ansible.plugins.lookup import LookupBase
from ansible import constants  # pylint: disable=unused-import


class LookupModule(LookupBase):
    def run(self, terms, variables, **kwargs):
        self.set_options(var_options=variables, direct=kwargs)

        return terms



from __future__ import annotations

import traceback

from ansible._internal._errors import _error_factory
from ansible._internal._event_formatting import format_event_traceback
from units.mock.error_helper import raise_exceptions

import pytest


def test_traceback_formatting() -> None:
    """Verify our traceback formatting mimics the Python traceback formatting."""
    with pytest.raises(Exception) as error:
        raise_exceptions((
            Exception('a'),
            Exception('b'),
            Exception('c'),
            Exception('d'),
        ))

    event = _error_factory.ControllerEventFactory.from_exception(error.value, True)
    ansible_tb = format_event_traceback(event)
    python_tb = ''.join(traceback.format_exception(error.value))

    assert ansible_tb == python_tb



from __future__ import annotations

import dataclasses
import typing as t

from ansible.template import accept_args_markers
from ansible._internal._templating._jinja_common import ExceptionMarker


@accept_args_markers
def dump_object(value: t.Any) -> object:
    """Internal filter to convert objects not supported by JSON to types which are."""
    if dataclasses.is_dataclass(value):
        return dataclasses.asdict(value)  # type: ignore[arg-type]

    if isinstance(value, ExceptionMarker):
        return dict(
            exception=value._as_exception(),
        )

    return value


class FilterModule(object):
    @staticmethod
    def filters() -> dict[str, t.Callable]:
        return dict(dump_object=dump_object)



# -*- coding: utf-8 -*-
# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = """
author: 'Ansible Core Team (@ansible)'
short_description: 'Add and remove deb822 formatted repositories'
description:
- 'Add and remove deb822 formatted repositories in Debian based distributions.'
module: deb822_repository
notes:
- This module will not automatically update caches, call the M(ansible.builtin.apt) module based
  on the changed state.
options:
    allow_downgrade_to_insecure:
        description:
        - Allow downgrading a package that was previously authenticated but
          is no longer authenticated.
        type: bool
    allow_insecure:
        description:
        - Allow insecure repositories.
        type: bool
    allow_weak:
        description:
        - Allow repositories signed with a key using a weak digest algorithm.
        type: bool
    architectures:
        description:
        - Architectures to search within repository.
        type: list
        elements: str
    by_hash:
        description:
        - Controls if APT should try to acquire indexes via a URI constructed
          from a hashsum of the expected file instead of using the well-known
          stable filename of the index.
        type: bool
    check_date:
        description:
        - Controls if APT should consider the machine's time correct and hence
          perform time related checks, such as verifying that a Release file
          is not from the future.
        type: bool
    check_valid_until:
        description:
        - Controls if APT should try to detect replay attacks.
        type: bool
    components:
        description:
        - Components specify different sections of one distribution version
          present in a C(Suite).
        type: list
        elements: str
    date_max_future:
        description:
        - Controls how far from the future a repository may be.
        type: int
    enabled:
        description:
        - Tells APT whether the source is enabled or not.
        type: bool
    inrelease_path:
        description:
        - Determines the path to the C(InRelease) file, relative to the normal
          position of an C(InRelease) file.
        type: str
    languages:
        description:
        - Defines which languages information such as translated
          package descriptions should be downloaded.
        type: list
        elements: str
    name:
        description:
        - Name of the repo. Specifically used for C(X-Repolib-Name) and in
          naming the repository and signing key files.
        required: true
        type: str
    pdiffs:
        description:
        - Controls if APT should try to use C(PDiffs) to update old indexes
          instead of downloading the new indexes entirely.
        type: bool
    signed_by:
        description:
        - Either a URL to a GPG key, absolute path to a keyring file, one or
          more fingerprints of keys either in the C(trusted.gpg) keyring or in
          the keyrings in the C(trusted.gpg.d/) directory, or an ASCII armored
          GPG public key block.
        type: str
    suites:
        description:
        - >-
          Suite can specify an exact path in relation to the URI(s) provided,
          in which case the Components: must be omitted and suite must end
          with a slash (C(/)). Alternatively, it may take the form of a
          distribution version (for example a version codename like C(disco) or C(artful)).
          If the suite does not specify a path, at least one component must
          be present.
        type: list
        elements: str
    targets:
        description:
        - Defines which download targets apt will try to acquire from this source.
        type: list
        elements: str
    trusted:
        description:
        - Decides if a source is considered trusted or if warnings should be
          raised before, for example packages are installed from this source.
        type: bool
    types:
        choices:
        - deb
        - deb-src
        default:
        - deb
        type: list
        elements: str
        description:
        - Which types of packages to look for from a given source; either
          binary V(deb) or source code V(deb-src).
    uris:
        description:
        - The URIs must specify the base of the Debian distribution archive,
          from which APT finds the information it needs.
        type: list
        elements: str
    mode:
        description:
        - The octal mode for newly created files in C(sources.list.d).
        type: raw
        default: '0644'
    state:
        description:
        - A source string state.
        type: str
        choices:
        - absent
        - present
        default: present
requirements:
    - python3-debian / python-debian
version_added: '2.15'
"""

EXAMPLES = """
- name: Add debian repo
  deb822_repository:
    name: debian
    types: deb
    uris: http://deb.debian.org/debian
    suites: stretch
    components:
      - main
      - contrib
      - non-free

- name: Add debian repo with key
  deb822_repository:
    name: debian
    types: deb
    uris: https://deb.debian.org
    suites: stable
    components:
      - main
      - contrib
      - non-free
    signed_by: |-
      -----BEGIN PGP PUBLIC KEY BLOCK-----

      mDMEYCQjIxYJKwYBBAHaRw8BAQdAD/P5Nvvnvk66SxBBHDbhRml9ORg1WV5CvzKY
      CuMfoIS0BmFiY2RlZoiQBBMWCgA4FiEErCIG1VhKWMWo2yfAREZd5NfO31cFAmAk
      IyMCGyMFCwkIBwMFFQoJCAsFFgIDAQACHgECF4AACgkQREZd5NfO31fbOwD6ArzS
      dM0Dkd5h2Ujy1b6KcAaVW9FOa5UNfJ9FFBtjLQEBAJ7UyWD3dZzhvlaAwunsk7DG
      3bHcln8DMpIJVXht78sL
      =IE0r
      -----END PGP PUBLIC KEY BLOCK-----

- name: Add repo using key from URL
  deb822_repository:
    name: example
    types: deb
    uris: https://download.example.com/linux/ubuntu
    suites: '{{ ansible_distribution_release }}'
    components: stable
    architectures: amd64
    signed_by: https://download.example.com/linux/ubuntu/gpg
"""

RETURN = """
repo:
  description: A source string for the repository
  returned: always
  type: str
  sample: |
    X-Repolib-Name: debian
    Types: deb
    URIs: https://deb.debian.org
    Suites: stable
    Components: main contrib non-free
    Signed-By:
        -----BEGIN PGP PUBLIC KEY BLOCK-----
        .
        mDMEYCQjIxYJKwYBBAHaRw8BAQdAD/P5Nvvnvk66SxBBHDbhRml9ORg1WV5CvzKY
        CuMfoIS0BmFiY2RlZoiQBBMWCgA4FiEErCIG1VhKWMWo2yfAREZd5NfO31cFAmAk
        IyMCGyMFCwkIBwMFFQoJCAsFFgIDAQACHgECF4AACgkQREZd5NfO31fbOwD6ArzS
        dM0Dkd5h2Ujy1b6KcAaVW9FOa5UNfJ9FFBtjLQEBAJ7UyWD3dZzhvlaAwunsk7DG
        3bHcln8DMpIJVXht78sL
        =IE0r
        -----END PGP PUBLIC KEY BLOCK-----

dest:
  description: Path to the repository file
  returned: always
  type: str
  sample: /etc/apt/sources.list.d/focal-archive.sources

key_filename:
  description: Path to the signed_by key file
  returned: always
  type: str
  sample: /etc/apt/keyrings/debian.gpg
"""

import os
import re
import tempfile
import textwrap

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.basic import missing_required_lib
from ansible.module_utils.common.collections import is_sequence
from ansible.module_utils.common.file import S_IRWXU_RXG_RXO, S_IRWU_RG_RO
from ansible.module_utils.common.text.converters import to_bytes
from ansible.module_utils.common.text.converters import to_native
from ansible.module_utils.six import raise_from  # type: ignore[attr-defined]
from ansible.module_utils.urls import generic_urlparse
from ansible.module_utils.urls import open_url
from ansible.module_utils.urls import get_user_agent
from ansible.module_utils.urls import urlparse

HAS_DEBIAN = True
DEBIAN_IMP_ERR = None
try:
    from debian.deb822 import Deb822  # type: ignore[import]
except ImportError as ex:
    HAS_DEBIAN = False
    DEBIAN_IMP_ERR = ex

KEYRINGS_DIR = '/etc/apt/keyrings'


def ensure_keyrings_dir(module):
    changed = False
    if not os.path.isdir(KEYRINGS_DIR):
        if not module.check_mode:
            os.mkdir(KEYRINGS_DIR, S_IRWXU_RXG_RXO)
        changed |= True

    changed |= module.set_fs_attributes_if_different(
        {
            'path': KEYRINGS_DIR,
            'secontext': [None, None, None],
            'owner': 'root',
            'group': 'root',
            'mode': '0755',
            'attributes': None,
        },
        changed,
    )

    return changed


def make_signed_by_filename(slug, ext):
    return os.path.join(KEYRINGS_DIR, '%s.%s' % (slug, ext))


def make_sources_filename(slug):
    return os.path.join(
        '/etc/apt/sources.list.d',
        '%s.sources' % slug
    )


def format_bool(v):
    return 'yes' if v else 'no'


def format_list(v):
    return ' '.join(v)


def format_multiline(v):
    return '\n' + textwrap.indent(
        '\n'.join(line.strip() or '.' for line in v.strip().splitlines()),
        '    '
    )


def format_field_name(v):
    if v == 'name':
        return 'X-Repolib-Name'
    elif v == 'uris':
        return 'URIs'
    return v.replace('_', '-').title()


def is_armored(b_data):
    return b'-----BEGIN PGP PUBLIC KEY BLOCK-----' in b_data


def write_signed_by_key(module, v, slug):
    changed = False
    if os.path.isfile(v):
        return changed, v, None

    b_data = None

    parts = generic_urlparse(urlparse(v))
    if parts.scheme:
        try:
            r = open_url(v, http_agent=get_user_agent())
        except Exception as exc:
            raise_from(RuntimeError(to_native(exc)), exc)
        else:
            b_data = r.read()
    else:
        # Not a file, nor a URL, just pass it through
        return changed, None, v

    if not b_data:
        return changed, v, None

    tmpfd, tmpfile = tempfile.mkstemp(dir=module.tmpdir)
    with os.fdopen(tmpfd, 'wb') as f:
        f.write(b_data)

    ext = 'asc' if is_armored(b_data) else 'gpg'
    filename = make_signed_by_filename(slug, ext)

    src_chksum = module.sha256(tmpfile)
    dest_chksum = module.sha256(filename)

    if src_chksum != dest_chksum:
        changed |= ensure_keyrings_dir(module)
        if not module.check_mode:
            module.atomic_move(tmpfile, filename)
        changed |= True

    changed |= module.set_mode_if_different(filename, S_IRWU_RG_RO, False)

    return changed, filename, None


def main():
    module = AnsibleModule(
        argument_spec={
            'allow_downgrade_to_insecure': {
                'type': 'bool',
            },
            'allow_insecure': {
                'type': 'bool',
            },
            'allow_weak': {
                'type': 'bool',
            },
            'architectures': {
                'elements': 'str',
                'type': 'list',
            },
            'by_hash': {
                'type': 'bool',
            },
            'check_date': {
                'type': 'bool',
            },
            'check_valid_until': {
                'type': 'bool',
            },
            'components': {
                'elements': 'str',
                'type': 'list',
            },
            'date_max_future': {
                'type': 'int',
            },
            'enabled': {
                'type': 'bool',
            },
            'inrelease_path': {
                'type': 'str',
            },
            'languages': {
                'elements': 'str',
                'type': 'list',
            },
            'name': {
                'type': 'str',
                'required': True,
            },
            'pdiffs': {
                'type': 'bool',
            },
            'signed_by': {
                'type': 'str',
            },
            'suites': {
                'elements': 'str',
                'type': 'list',
            },
            'targets': {
                'elements': 'str',
                'type': 'list',
            },
            'trusted': {
                'type': 'bool',
            },
            'types': {
                'choices': [
                    'deb',
                    'deb-src',
                ],
                'elements': 'str',
                'type': 'list',
                'default': [
                    'deb',
                ]
            },
            'uris': {
                'elements': 'str',
                'type': 'list',
            },
            # non-deb822 args
            'mode': {
                'type': 'raw',
                'default': '0644',
            },
            'state': {
                'type': 'str',
                'choices': [
                    'present',
                    'absent',
                ],
                'default': 'present',
            },
        },
        supports_check_mode=True,
    )

    if not HAS_DEBIAN:
        module.fail_json(msg=missing_required_lib("python3-debian"),
                         exception=DEBIAN_IMP_ERR)

    check_mode = module.check_mode

    changed = False

    # Make a copy, so we don't mutate module.params to avoid future issues
    params = module.params.copy()

    # popped non-deb822 args
    mode = params.pop('mode')
    state = params.pop('state')

    name = params['name']
    slug = re.sub(
        r'[^a-z0-9-]+',
        '',
        re.sub(
            r'[_\s]+',
            '-',
            name.lower(),
        ),
    )
    sources_filename = make_sources_filename(slug)

    if state == 'absent':
        if os.path.exists(sources_filename):
            if not check_mode:
                os.unlink(sources_filename)
            changed |= True
        for ext in ('asc', 'gpg'):
            signed_by_filename = make_signed_by_filename(slug, ext)
            if os.path.exists(signed_by_filename):
                if not check_mode:
                    os.unlink(signed_by_filename)
                changed = True
        module.exit_json(
            repo=None,
            changed=changed,
            dest=sources_filename,
            key_filename=signed_by_filename,
        )

    deb822 = Deb822()
    signed_by_filename = None
    for key, value in sorted(params.items()):
        if value is None:
            continue

        if isinstance(value, bool):
            value = format_bool(value)
        elif isinstance(value, int):
            value = to_native(value)
        elif is_sequence(value):
            value = format_list(value)
        elif key == 'signed_by':
            try:
                key_changed, signed_by_filename, signed_by_data = write_signed_by_key(module, value, slug)
                value = signed_by_filename or signed_by_data
                changed |= key_changed
            except RuntimeError as exc:
                module.fail_json(
                    msg='Could not fetch signed_by key: %s' % to_native(exc)
                )

        if value.count('\n') > 0:
            value = format_multiline(value)

        deb822[format_field_name(key)] = value

    repo = deb822.dump()
    tmpfd, tmpfile = tempfile.mkstemp(dir=module.tmpdir)
    with os.fdopen(tmpfd, 'wb') as f:
        f.write(to_bytes(repo))

    sources_filename = make_sources_filename(slug)

    src_chksum = module.sha256(tmpfile)
    dest_chksum = module.sha256(sources_filename)

    if src_chksum != dest_chksum:
        if not check_mode:
            module.atomic_move(tmpfile, sources_filename)
        changed |= True

    changed |= module.set_mode_if_different(sources_filename, mode, False)

    module.exit_json(
        repo=repo,
        changed=changed,
        dest=sources_filename,
        key_filename=signed_by_filename,
    )


if __name__ == '__main__':
    main()



"""
This file is not used by the integration test, but serves a related purpose.
It triggers sanity test failures that can only occur for ansible-core, which need to be ignored to ensure the pylint plugin is functioning properly.
"""
from __future__ import annotations

from ansible.module_utils.datatag import deprecator_from_collection_name
from ansible.module_utils.common.warnings import deprecate


def do_stuff() -> None:
    deprecator = deprecator_from_collection_name('ansible.builtin')

    deprecate(msg='ansible-deprecated-version', version='2.18')
    deprecate(msg='ansible-deprecated-no-version')
    deprecate(msg='ansible-invalid-deprecated-version', version='not-a-version')
    # collection-deprecated-version - ansible-core cannot encounter this
    # collection-invalid-deprecated-version - ansible-core cannot encounter this
    # ansible-deprecated-no-collection-name - ansible-core cannot encounter this
    # wrong-collection-deprecated - ansible-core cannot encounter this
    # ansible-expired-deprecated-date - ansible-core cannot encounter this
    # ansible-invalid-deprecated-date - ansible-core cannot encounter this
    # ansible-deprecated-both-version-and-date - ansible-core cannot encounter this
    # removal-version-must-be-major - ansible-core cannot encounter this
    deprecate(msg='ansible-deprecated-date-not-permitted', date='2099-01-01')
    deprecate(msg='ansible-deprecated-unnecessary-collection-name', deprecator=deprecator, version='2.99')
    deprecate(msg='ansible-deprecated-collection-name-not-permitted', collection_name='ansible.builtin', version='2.99')
    # ansible-deprecated-both-collection-name-and-deprecator - ansible-core cannot encounter this



# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations


class ModuleDocFragment(object):
    DOCUMENTATION = r"""
options:
    testcol2option:
        description:
            - An option taken from testcol2
        type: str
        version_added: 1.0.0
    testcol2option2:
        description:
            - Another option taken from testcol2
        type: str
"""



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations


import pytest

from unittest.mock import Mock


@pytest.fixture
def mock_module():
    mock_module = Mock()
    mock_module.params = {'gather_subset': ['all'],
                          'gather_timeout': 5,
                          'filter': '*'}
    mock_module.get_bin_path = Mock(return_value=None)
    return mock_module



#
# (c) 2020 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import sys
from io import StringIO
import pytest
import typing as t

from ansible.module_utils import _internal
from ansible.plugins.connection import paramiko_ssh as paramiko_ssh_module
from ansible.plugins.loader import connection_loader
from ansible.playbook.play_context import PlayContext

from units.test_utils.controller.display import emits_warnings


@pytest.fixture
def play_context():
    play_context = PlayContext()
    play_context.prompt = (
        '[sudo via ansible, key=ouzmdnewuhucvuaabtjmweasarviygqq] password: '
    )

    return play_context


@pytest.fixture()
def in_stream():
    return StringIO()


def test_paramiko_connection_module(play_context, in_stream):
    assert isinstance(
        connection_loader.get('paramiko_ssh', play_context, in_stream),
        paramiko_ssh_module.Connection)


def test_paramiko_connect(play_context, in_stream, mocker):
    paramiko_ssh = connection_loader.get('paramiko_ssh', play_context, in_stream)
    mocker.patch.object(paramiko_ssh, '_connect_uncached')
    connection = paramiko_ssh._connect()

    assert isinstance(connection, paramiko_ssh_module.Connection)
    assert connection._connected is True


def test_deprecation_warning_controller(monkeypatch: pytest.MonkeyPatch) -> None:
    """Ensures deprecation warnings are generated for external paramiko imports."""
    assert _internal.is_controller

    sentinel: t.Any = object()

    monkeypatch.delitem(sys.modules, 'ansible')
    monkeypatch.delitem(sys.modules, 'ansible.module_utils')
    monkeypatch.delitem(sys.modules, 'ansible.module_utils.compat')
    monkeypatch.delitem(sys.modules, 'ansible.module_utils.compat.paramiko')
    monkeypatch.setitem(sys.modules, 'paramiko', sentinel)

    # ensure direct access to `_` prefixed attrs does not warn
    with emits_warnings(deprecation_pattern=[], warning_pattern=[]):
        from ansible.module_utils.compat import paramiko
        assert paramiko._paramiko is sentinel
        assert isinstance(paramiko._PARAMIKO_IMPORT_ERR, (Exception, type(None)))

    with emits_warnings(deprecation_pattern=["The 'paramiko' compat import is deprecated", "The 'PARAMIKO_IMPORT_ERR' compat import is deprecated"]):
        from ansible.module_utils.compat import paramiko
        assert paramiko.paramiko is sentinel
        assert isinstance(paramiko.PARAMIKO_IMPORT_ERR, (Exception, type(None)))



from __future__ import annotations

import sys
import pytest
import subprocess

from ansible.module_utils import basic
from ansible.module_utils.testing import patch_module_args


def test_run_non_existent_command(monkeypatch):
    """ Test that `command` returns std{out,err} even if the executable is not found """
    def fail_json(msg, **kwargs):
        assert kwargs["stderr"] == ''
        assert kwargs["stdout"] == ''
        sys.exit(1)

    def popen(*args, **kwargs):
        raise OSError()

    monkeypatch.setattr(subprocess, 'Popen', popen)

    with patch_module_args():
        am = basic.AnsibleModule(argument_spec={})

    monkeypatch.setattr(am, 'fail_json', fail_json)
    with pytest.raises(SystemExit):
        am.run_command("lecho", "whatever")



from __future__ import annotations

import typing as t

from ansible.template import accept_args_markers


@accept_args_markers
def true_type(obj: object) -> str:
    """Internal filter to show the true type name of the given object, not just the base type name like the `debug` filter."""
    return obj.__class__.__name__


class FilterModule(object):
    @staticmethod
    def filters() -> dict[str, t.Callable]:
        return dict(true_type=true_type)



# -*- coding: utf-8 -*-
# (c) 2018, Ansible Project
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
"""
This test checks whether the libraries we're bundling are out of date and need to be synced with
a newer upstream release.
"""


from __future__ import annotations

import fnmatch
import json
import re
import sys
from ansible.module_utils.compat.version import LooseVersion

import packaging.specifiers

from ansible.module_utils.urls import open_url


BUNDLED_RE = re.compile(b'\\b_BUNDLED_METADATA\\b')


def get_bundled_libs(paths):
    """
    Return the set of known bundled libraries

    :arg paths: The paths which the test has been instructed to check
    :returns: The list of all files which we know to contain bundled libraries.  If a bundled
        library consists of multiple files, this should be the file which has metadata included.
    """
    bundled_libs = set()
    for filename in fnmatch.filter(paths, 'lib/ansible/compat/*/__init__.py'):
        bundled_libs.add(filename)

    bundled_libs.add('lib/ansible/module_utils/distro/__init__.py')
    bundled_libs.add('lib/ansible/module_utils/six/__init__.py')

    return bundled_libs


def get_files_with_bundled_metadata(paths):
    """
    Search for any files which have bundled metadata inside of them

    :arg paths: Iterable of filenames to search for metadata inside of
    :returns: A set of pathnames which contained metadata
    """

    with_metadata = set()
    for path in paths:
        with open(path, 'rb') as f:
            body = f.read()

        if BUNDLED_RE.search(body):
            with_metadata.add(path)

    return with_metadata


def get_bundled_metadata(filename):
    """
    Retrieve the metadata about a bundled library from a python file

    :arg filename: The filename to look inside for the metadata
    :raises ValueError: If we're unable to extract metadata from the file
    :returns: The metadata from the python file
    """
    with open(filename, 'r') as module:
        for line in module:
            if line.strip().startswith('# NOT_BUNDLED'):
                return None

            if line.strip().startswith('# CANT_UPDATE'):
                print(f'{filename} marked as CANT_UPDATE, so skipping. Manual check for CVEs required.')
                return None

            if line.strip().startswith('_BUNDLED_METADATA'):
                data = line[line.index('{') :].strip()
                break
        else:
            raise ValueError(
                'Unable to check bundled library for update. '
                'Please add _BUNDLED_METADATA dictionary to the library file with information on pypi name and bundled version.'
            )
        metadata = json.loads(data)
    return metadata


def get_latest_applicable_version(pypi_data, constraints=None):
    """Get the latest pypi version of the package that we allow

    :arg pypi_data: Pypi information about the data as returned by
        ``https://pypi.org/pypi/{pkg_name}/json``
    :kwarg constraints: version constraints on what we're allowed to use as specified by
        the bundled metadata
    :returns: The most recent version on pypi that are allowed by ``constraints``
    """
    latest_version = "0"
    if constraints:
        version_specification = packaging.specifiers.SpecifierSet(constraints)
        for version in pypi_data['releases']:
            if version in version_specification:
                if LooseVersion(version) > LooseVersion(latest_version):
                    latest_version = version
    else:
        latest_version = pypi_data['info']['version']

    return latest_version


def main():
    """Entrypoint to the script"""

    paths = sys.argv[1:] or sys.stdin.read().splitlines()

    bundled_libs = get_bundled_libs(paths)
    files_with_bundled_metadata = get_files_with_bundled_metadata(paths)

    for filename in files_with_bundled_metadata.difference(bundled_libs):
        if filename.startswith('test/support/'):
            continue  # bundled support code does not need to be updated or tracked

        print(f'{filename}: ERROR: File contains _BUNDLED_METADATA but needs to be added to test/sanity/code-smell/update-bundled.py')

    for filename in bundled_libs:
        try:
            metadata = get_bundled_metadata(filename)
        except ValueError as e:
            print(f'{filename}: ERROR: {e}')
            continue
        except OSError as e:
            if e.errno == 2:
                print(
                    f'{filename}: ERROR: {e}. '
                    'Perhaps the bundled library has been removed or moved and the bundled library test needs to be modified as well?'
                )

        if metadata is None:
            continue

        pypi_fh = open_url('https://pypi.org/pypi/{0}/json'.format(metadata['pypi_name']))
        pypi_data = json.loads(pypi_fh.read().decode('utf-8'))

        constraints = metadata.get('version_constraints', None)
        latest_version = get_latest_applicable_version(pypi_data, constraints)

        if LooseVersion(metadata['version']) < LooseVersion(latest_version):
            name = metadata['pypi_name']
            version = metadata['version']
            url = f"https://pypi.org/pypi/{name}/json"

            print(f"{filename}: UPDATE {name} from {version} to {latest_version} {url}")


if __name__ == '__main__':
    main()



#!/usr/bin/python
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = """
module: module1
short_description: Hello test module
description: Hello test module.
options: {}
author:
  - Ansible Core Team
"""

EXAMPLES = """
- minimal:
"""

RETURN = """"""

from ansible.module_utils.basic import AnsibleModule


def main():
    module = AnsibleModule(
        argument_spec={},
    )

    module.exit_json()


if __name__ == '__main__':
    main()



# Copyright (c) 2024 Ansible Project
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)

"""Code generation of __post_init__ methods for efficient dataclass field type checking at runtime."""

from __future__ import annotations

import atexit
import functools
import itertools
import shutil
import tempfile
import types
import typing as t

_write_generated_code_to_disk = False

# deprecated: description='types.UnionType is available in Python 3.10' python_version='3.9'
try:
    _union_type: type | None = types.UnionType  # type: ignore[attr-defined]
    _union_types: tuple = (t.Union, types.UnionType)  # type: ignore[attr-defined]
except AttributeError:
    _union_type = None  # type: ignore[assignment]
    _union_types = (t.Union,)  # type: ignore[assignment]


def inject_post_init_validation(cls: type, allow_subclasses=False) -> None:
    """Inject a __post_init__ field validation method on the given dataclass. An existing __post_init__ attribute must already exist."""
    # DTFIX-FUTURE: when cls must have a __post_init__, enforcing it as a no-op would be nice, but is tricky on slotted dataclasses due to double-creation
    post_validate_name = '_post_validate'
    method_name = '__post_init__'
    exec_globals: dict[str, t.Any] = {}
    known_types: dict[type, str] = {}
    lines: list[str] = []
    field_type_hints = t.get_type_hints(cls)
    indent = 1

    def append_line(line: str) -> None:
        """Append a line to the generated source at the current indentation level."""
        lines.append((' ' * indent * 4) + line)

    def register_type(target_type: type) -> str:
        """Register the target type and return the local name."""
        target_name = f'{target_type.__module__.replace(".", "_")}_{target_type.__name__}'

        known_types[target_type] = target_name
        exec_globals[target_name] = target_type

        return target_name

    def validate_value(target_name: str, target_ref: str, target_type: type) -> None:
        """Generate code to validate the specified value."""
        nonlocal indent

        origin_type = t.get_origin(target_type)

        if origin_type is t.ClassVar:
            return  # ignore annotations which are not fields, indicated by the t.ClassVar annotation

        allowed_types = _get_allowed_types(target_type)

        # check value

        if origin_type is t.Literal:
            # DTFIX-FUTURE: support optional literals

            values = t.get_args(target_type)

            append_line(f"""if {target_ref} not in {values}:""")
            append_line(f"""    raise ValueError(rf"{target_name} must be one of {values} instead of {{{target_ref}!r}}")""")

        allowed_refs = [register_type(allowed_type) for allowed_type in allowed_types]
        allowed_names = [repr(allowed_type) for allowed_type in allowed_types]

        if allow_subclasses:
            if len(allowed_refs) == 1:
                append_line(f"""if not isinstance({target_ref}, {allowed_refs[0]}):""")
            else:
                append_line(f"""if not isinstance({target_ref}, ({', '.join(allowed_refs)})):""")
        else:
            if len(allowed_refs) == 1:
                append_line(f"""if type({target_ref}) is not {allowed_refs[0]}:""")
            else:
                append_line(f"""if type({target_ref}) not in ({', '.join(allowed_refs)}):""")

        append_line(f"""    raise TypeError(f"{target_name} must be {' or '.join(allowed_names)} instead of {{type({target_ref})}}")""")

        # check elements (for containers)

        if target_ref.startswith('self.'):
            local_ref = target_ref[5:]
        else:
            local_ref = target_ref

        if tuple in allowed_types:
            tuple_type = _extract_type(target_type, tuple)

            idx_ref = f'{local_ref}_idx'
            item_ref = f'{local_ref}_item'
            item_name = f'{target_name}[{{{idx_ref}!r}}]'
            item_type, _ellipsis = t.get_args(tuple_type)

            if _ellipsis is not ...:
                raise ValueError(f"{cls} tuple fields must be a tuple of a single element type")

            append_line(f"""if isinstance({target_ref}, {known_types[tuple]}):""")
            append_line(f"""    for {idx_ref}, {item_ref} in enumerate({target_ref}):""")

            indent += 2
            validate_value(target_name=item_name, target_ref=item_ref, target_type=item_type)
            indent -= 2

        if list in allowed_types:
            list_type = _extract_type(target_type, list)

            idx_ref = f'{local_ref}_idx'
            item_ref = f'{local_ref}_item'
            item_name = f'{target_name}[{{{idx_ref}!r}}]'
            (item_type,) = t.get_args(list_type)

            append_line(f"""if isinstance({target_ref}, {known_types[list]}):""")
            append_line(f"""    for {idx_ref}, {item_ref} in enumerate({target_ref}):""")

            indent += 2
            validate_value(target_name=item_name, target_ref=item_ref, target_type=item_type)
            indent -= 2

        if dict in allowed_types:
            dict_type = _extract_type(target_type, dict)

            key_ref, value_ref = f'{local_ref}_key', f'{local_ref}_value'
            key_type, value_type = t.get_args(dict_type)
            key_name, value_name = f'{target_name!r} key {{{key_ref}!r}}', f'{target_name}[{{{key_ref}!r}}]'

            append_line(f"""if isinstance({target_ref}, {known_types[dict]}):""")
            append_line(f"""    for {key_ref}, {value_ref} in {target_ref}.items():""")

            indent += 2
            validate_value(target_name=key_name, target_ref=key_ref, target_type=key_type)
            validate_value(target_name=value_name, target_ref=value_ref, target_type=value_type)
            indent -= 2

    for field_name in cls.__annotations__:
        validate_value(target_name=f'{{type(self).__name__}}.{field_name}', target_ref=f'self.{field_name}', target_type=field_type_hints[field_name])

    if hasattr(cls, post_validate_name):
        append_line(f"self.{post_validate_name}()")

    if not lines:
        return  # nothing to validate (empty dataclass)

    if '__init__' in cls.__dict__ and not hasattr(cls, method_name):
        raise ValueError(f"{cls} must have a {method_name!r} method to override when invoked after the '__init__' method is created")

    if any(hasattr(parent, method_name) for parent in cls.__mro__[1:]):
        lines.insert(0, f'    super({register_type(cls)}, self).{method_name}()')

    lines.insert(0, f'def {method_name}(self):')

    source = '\n'.join(lines) + '\n'

    if _write_generated_code_to_disk:
        tmp = tempfile.NamedTemporaryFile(mode='w+t', suffix=f'-{cls.__module__}.{cls.__name__}.py', delete=False, dir=_get_temporary_directory())

        tmp.write(source)
        tmp.flush()

        filename = tmp.name
    else:
        filename = f'<string> generated for {cls}'

    code = compile(source, filename, 'exec')

    exec(code, exec_globals)
    setattr(cls, method_name, exec_globals[method_name])


@functools.lru_cache(maxsize=1)
def _get_temporary_directory() -> str:
    """Create a temporary directory and return its full path. The directory will be deleted when the process exits."""
    temp_dir = tempfile.mkdtemp()

    atexit.register(lambda: shutil.rmtree(temp_dir))

    return temp_dir


def _get_allowed_types(target_type: type) -> tuple[type, ...]:
    """Return a tuple of types usable in instance checks for the given target_type."""
    origin_type = t.get_origin(target_type)

    if origin_type in _union_types:
        allowed_types = tuple(set(itertools.chain.from_iterable(_get_allowed_types(arg) for arg in t.get_args(target_type))))
    elif origin_type is t.Literal:
        allowed_types = (str,)  # DTFIX-FUTURE: support non-str literal types
    elif origin_type:
        allowed_types = (origin_type,)
    else:
        allowed_types = (target_type,)

    return allowed_types


def _extract_type(target_type: type, of_type: type) -> type:
    """Return `of_type` from `target_type`, where `target_type` may be a union."""
    origin_type = t.get_origin(target_type)

    if origin_type is of_type:  # pylint: disable=unidiomatic-typecheck
        return target_type

    if origin_type is t.Union or (_union_type and isinstance(target_type, _union_type)):
        args = t.get_args(target_type)
        extracted_types = [arg for arg in args if type(arg) is of_type or t.get_origin(arg) is of_type]  # pylint: disable=unidiomatic-typecheck
        (extracted_type,) = extracted_types
        return extracted_type

    raise NotImplementedError(f'{target_type} is not supported')



from __future__ import annotations

from ansible.module_utils.basic import _load_params, AnsibleModule
from ansible.module_utils.testing import patch_module_args


def do_echo():
    p = _load_params()

    with patch_module_args():
        module = AnsibleModule(argument_spec={})
        module.exit_json(args_in=p)



# -*- coding: utf-8 -*-
# Copyright: (c) 2020-2021, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""Requirement reporter implementations."""

from __future__ import annotations

from collections import defaultdict

try:
    from resolvelib import BaseReporter
except ImportError:
    class BaseReporter:  # type: ignore[no-redef]
        pass

try:
    from resolvelib.resolvers import Criterion
except ImportError:
    class Criterion:  # type: ignore[no-redef]
        pass

from ansible.utils.display import Display
from .dataclasses import Candidate, Requirement


display = Display()


_CLI_APP_NAME = 'ansible-galaxy'
_MESSAGES_AT_REJECT_COUNT = {
    1: (
        f'{_CLI_APP_NAME} is looking at multiple versions of {{fqcn}} to '
        'determine which version is compatible with other '
        'requirements. This could take a while.'
    ),
    8: (
        f'{_CLI_APP_NAME} is looking at multiple versions of {{fqcn}} to '
        'determine which version is compatible with other '
        'requirements. This could take a while.'
    ),
    13: (
        'This is taking longer than usual. You might need to provide '
        'the dependency resolver with stricter constraints to reduce '
        'runtime. If you want to abort this run, press Ctrl + C.'
    ),
}


class CollectionDependencyReporter(BaseReporter):
    """A dependency reporter for Ansible Collections.

    This is a proxy class allowing us to abstract away importing resolvelib
    outside of the `ansible.galaxy.dependency_resolution` Python package.
    """

    def __init__(self) -> None:
        """Initialize the collection rejection counter."""
        super().__init__()

        self.reject_count_by_fqcn: defaultdict[str, int] = defaultdict(int)

    def _maybe_log_rejection_message(self, candidate: Candidate) -> bool:
        """Print out rejection messages on pre-defined limit hits."""
        # Inspired by https://github.com/pypa/pip/commit/9731131
        self.reject_count_by_fqcn[candidate.fqcn] += 1

        collection_rejections_count = self.reject_count_by_fqcn[candidate.fqcn]

        if collection_rejections_count not in _MESSAGES_AT_REJECT_COUNT:
            return False

        collection_rejection_message = _MESSAGES_AT_REJECT_COUNT[
            collection_rejections_count
        ]
        display.display(collection_rejection_message.format(fqcn=candidate.fqcn))

        return True

    def rejecting_candidate(  # resolvelib >= 0.9.0
            self,
            criterion: Criterion[Candidate, Requirement],
            candidate: Candidate,
    ) -> None:
        """Augment rejection messages with conflict details."""
        if not self._maybe_log_rejection_message(candidate):
            return

        msg = 'Will try a different candidate, due to conflict:'
        for req_info in criterion.information:
            req, parent = req_info.requirement, req_info.parent
            msg += '\n    '
            if parent:
                msg += f'{parent !s} depends on '
            else:
                msg += 'The user requested '
            msg += str(req)
        display.v(msg)

    def backtracking(self, candidate: Candidate) -> None:  # resolvelib < 0.9.0
        """Print out rejection messages on pre-defined limit hits."""
        self._maybe_log_rejection_message(candidate)



"""
Lossy best-effort serialization for Ansible variables; used primarily for callback JSON display.
Any type which is not supported by JSON will be converted to a string.
The string representation of any type that is not native to JSON is subject to change and should not be considered stable.
The decoder provides no special behavior.
"""

from __future__ import annotations as _annotations

import datetime as _datetime
import typing as _t

from json import dumps as _dumps

from ... import _datatag
from .. import _profiles


class _Profile(_profiles._JSONSerializationProfile["Encoder", "Decoder"]):
    serialize_map: _t.ClassVar[dict[type, _t.Callable]]

    @classmethod
    def post_init(cls) -> None:
        cls.serialize_map = {
            bytes: cls.serialize_bytes_as_str,
            set: cls.serialize_as_list,
            tuple: cls.serialize_as_list,
            _datetime.date: cls.serialize_as_isoformat,
            _datetime.time: cls.serialize_as_isoformat,
            _datetime.datetime: cls.serialize_as_isoformat,
            _datatag._AnsibleTaggedDate: cls.discard_tags,
            _datatag._AnsibleTaggedTime: cls.discard_tags,
            _datatag._AnsibleTaggedDateTime: cls.discard_tags,
            _datatag._AnsibleTaggedStr: cls.discard_tags,
            _datatag._AnsibleTaggedInt: cls.discard_tags,
            _datatag._AnsibleTaggedFloat: cls.discard_tags,
            _datatag._AnsibleTaggedSet: cls.discard_tags,
            _datatag._AnsibleTaggedList: cls.discard_tags,
            _datatag._AnsibleTaggedTuple: cls.discard_tags,
            _datatag._AnsibleTaggedDict: cls.discard_tags,
            _datatag._AnsibleTaggedBytes: cls.discard_tags,
        }

    @classmethod
    def serialize_bytes_as_str(cls, value: bytes) -> str:
        return value.decode(errors='surrogateescape')

    @classmethod
    def handle_key(cls, k: _t.Any) -> _t.Any:
        while mapped_callable := cls.serialize_map.get(type(k)):
            k = mapped_callable(k)

        k = cls.default(k)

        if not isinstance(k, str):
            k = _dumps(k, cls=Encoder)

        return k

    @classmethod
    def last_chance(cls, o: _t.Any) -> _t.Any:
        try:
            return str(o)
        except Exception as ex:
            return str(ex)


class Encoder(_profiles.AnsibleProfileJSONEncoder):
    _profile = _Profile


class Decoder(_profiles.AnsibleProfileJSONDecoder):
    _profile = _Profile



# Copyright 2015 Abhijit Menon-Sen <ams@2ndQuadrant.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import re
from ansible.errors import AnsibleParserError, AnsibleError

# Components that match a numeric or alphanumeric begin:end or begin:end:step
# range expression inside square brackets.

numeric_range = r"""
    \[
        (?:[0-9]+:[0-9]+)               # numeric begin:end
        (?::[0-9]+)?                    # numeric :step (optional)
    \]
"""

hexadecimal_range = r"""
    \[
        (?:[0-9a-f]+:[0-9a-f]+)         # hexadecimal begin:end
        (?::[0-9]+)?                    # numeric :step (optional)
    \]
"""

alphanumeric_range = r"""
    \[
        (?:
            [a-z]:[a-z]|                # one-char alphabetic range
            [0-9]+:[0-9]+               # ...or a numeric one
        )
        (?::[0-9]+)?                    # numeric :step (optional)
    \]
"""

# Components that match a 16-bit portion of an IPv6 address in hexadecimal
# notation (0..ffff) or an 8-bit portion of an IPv4 address in decimal notation
# (0..255) or an [x:y(:z)] numeric range.

ipv6_component = r"""
    (?:
        [0-9a-f]{{1,4}}|                # 0..ffff
        {range}                         # or a numeric range
    )
""".format(range=hexadecimal_range)

ipv4_component = r"""
    (?:
        [01]?[0-9]{{1,2}}|              # 0..199
        2[0-4][0-9]|                    # 200..249
        25[0-5]|                        # 250..255
        {range}                         # or a numeric range
    )
""".format(range=numeric_range)

# A hostname label, e.g. 'foo' in 'foo.example.com'. Consists of alphanumeric
# characters plus dashes (and underscores) or valid ranges. The label may not
# start or end with a hyphen or an underscore. This is interpolated into the
# hostname pattern below. We don't try to enforce the 63-char length limit.

label = r"""
    (?:[\w]|{range})                    # Starts with an alphanumeric or a range
    (?:[\w_-]|{range})*                 # Then zero or more of the same or [_-]
    (?<![_-])                           # ...as long as it didn't end with [_-]
""".format(range=alphanumeric_range)

patterns = {
    # This matches a square-bracketed expression with a port specification. What
    # is inside the square brackets is validated later.

    'bracketed_hostport': re.compile(
        r"""^
            \[(.+)\]                    # [host identifier]
            :([0-9]+)                   # :port number
            $
        """, re.X
    ),

    # This matches a bare IPv4 address or hostname (or host pattern including
    # [x:y(:z)] ranges) with a port specification.

    'hostport': re.compile(
        r"""^
            ((?:                        # We want to match:
                [^:\[\]]                # (a non-range character
                |                       # ...or...
                \[[^\]]*\]              # a complete bracketed expression)
            )*)                         # repeated as many times as possible
            :([0-9]+)                   # followed by a port number
            $
        """, re.X
    ),

    # This matches an IPv4 address, but also permits range expressions.

    'ipv4': re.compile(
        r"""^
            (?:{i4}\.){{3}}{i4}         # Three parts followed by dots plus one
            $
        """.format(i4=ipv4_component), re.X | re.I
    ),

    # This matches an IPv6 address, but also permits range expressions.
    #
    # This expression looks complex, but it really only spells out the various
    # combinations in which the basic unit of an IPv6 address (0..ffff) can be
    # written, from :: to 1:2:3:4:5:6:7:8, plus the IPv4-in-IPv6 variants such
    # as ::ffff:192.0.2.3.
    #
    # Note that we can't just use ipaddress.ip_address() because we also have to
    # accept ranges in place of each component.

    'ipv6': re.compile(
        r"""^
            ((?:{0}:){{7}}{0}|          # uncompressed: 1:2:3:4:5:6:7:8
            (?:{0}:){{1,6}}:|           # compressed variants, which are all
            (?:{0}:)(?::{0}){{1,6}}|    # a::b for various lengths of a,b
            (?:{0}:){{2}}(?::{0}){{1,5}}|
            (?:{0}:){{3}}(?::{0}){{1,4}}|
            (?:{0}:){{4}}(?::{0}){{1,3}}|
            (?:{0}:){{5}}(?::{0}){{1,2}}|
            (?:{0}:){{6}}(?::{0})|      # ...all with 2 <= a+b <= 7
            :(?::{0}){{1,6}}|           # ::ffff(:ffff...)
            {0}?::|                     # ffff::, ::
                                        # ipv4-in-ipv6 variants
            (?:0:){{6}}(?:{0}\.){{3}}{0}|
            ::(?:ffff:)?(?:{0}\.){{3}}{0}|
            (?:0:){{5}}ffff:(?:{0}\.){{3}}{0})
            $
        """.format(ipv6_component), re.X | re.I
    ),

    # This matches a hostname or host pattern including [x:y(:z)] ranges.
    #
    # We roughly follow DNS rules here, but also allow ranges (and underscores).
    # In the past, no systematic rules were enforced about inventory hostnames,
    # but the parsing context (e.g. shlex.split(), fnmatch.fnmatch()) excluded
    # various metacharacters anyway.
    #
    # We don't enforce DNS length restrictions here (63 characters per label,
    # 253 characters total) or make any attempt to process IDNs.

    'hostname': re.compile(
        r"""^
            {label}                     # We must have at least one label
            (?:\.{label})*              # Followed by zero or more .labels
            $
        """.format(label=label), re.X | re.I | re.UNICODE
    ),

}


def parse_address(address, allow_ranges=False):
    """
    Takes a string and returns a (host, port) tuple. If the host is None, then
    the string could not be parsed as a host identifier with an optional port
    specification. If the port is None, then no port was specified.

    The host identifier may be a hostname (qualified or not), an IPv4 address,
    or an IPv6 address. If allow_ranges is True, then any of those may contain
    [x:y] range specifications, e.g. foo[1:3] or foo[0:5]-bar[x-z].

    The port number is an optional :NN suffix on an IPv4 address or host name,
    or a mandatory :NN suffix on any square-bracketed expression: IPv6 address,
    IPv4 address, or host name. (This means the only way to specify a port for
    an IPv6 address is to enclose it in square brackets.)
    """

    # First, we extract the port number if one is specified.

    port = None
    for matching in ['bracketed_hostport', 'hostport']:
        m = patterns[matching].match(address)
        if m:
            (address, port) = m.groups()
            port = int(port)
            continue

    # What we're left with now must be an IPv4 or IPv6 address, possibly with
    # numeric ranges, or a hostname with alphanumeric ranges.

    host = None
    for matching in ['ipv4', 'ipv6', 'hostname']:
        m = patterns[matching].match(address)
        if m:
            host = address
            continue

    # If it isn't any of the above, we don't understand it.
    if not host:
        raise AnsibleError("Not a valid network hostname: %s" % address)

    # If we get to this point, we know that any included ranges are valid.
    # If the caller is prepared to handle them, all is well.
    # Otherwise we treat it as a parse failure.
    if not allow_ranges and '[' in host:
        raise AnsibleParserError("Detected range in host but was asked to ignore ranges")

    return (host, port)



"""Verify the contents of the built sdist and wheel."""

from __future__ import annotations

import contextlib
import fnmatch
import os
import pathlib
import re
import shutil
import subprocess
import sys
import tarfile
import tempfile
import typing as t
import zipfile

from ansible.release import __version__

fuzzy_match_basenames: set[str] = {'COPYING'}
"""Account for PEP639 differences in the placement of some files."""


def collect_sdist_files(complete_file_list: list[str]) -> list[str]:
    """Return a list of files which should be present in the sdist."""
    ignore_patterns = (
        '.azure-pipelines/*',
        '.cherry_picker.toml',
        '.git*',
        '.mailmap',
        'bin/*',
        'changelogs/README.md',
        'changelogs/config.yaml',
        'changelogs/fragments/*',
        'hacking/*',
    )

    sdist_files = [path for path in complete_file_list if not any(fnmatch.fnmatch(path, ignore) for ignore in ignore_patterns)]

    egg_info = (
        'PKG-INFO',
        'SOURCES.txt',
        'dependency_links.txt',
        'entry_points.txt',
        'requires.txt',
        'top_level.txt',
    )

    sdist_files.append('PKG-INFO')
    sdist_files.append('setup.cfg')
    sdist_files.extend(f'ansible_core.egg-info/{name}' for name in egg_info)

    return sdist_files


def collect_wheel_files(complete_file_list: list[str]) -> list[str]:
    """Return a list of files which should be present in the wheel."""
    wheel_files = []
    license_files = []

    for path in complete_file_list:
        if path.startswith('licenses/'):
            license_files.append(os.path.relpath(path, 'licenses'))
            fuzzy_match_basenames.add(os.path.basename(path))

        if path.startswith('lib/ansible/'):
            prefix = 'lib'
        elif path.startswith('test/lib/ansible_test/'):
            prefix = 'test/lib'
        else:
            continue

        wheel_files.append(os.path.relpath(path, prefix))

    dist_info = [
        'COPYING',
        'METADATA',
        'RECORD',
        'WHEEL',
        'entry_points.txt',
        'top_level.txt',
    ] + license_files

    wheel_files.extend(f'ansible_core-{__version__}.dist-info/{name}' for name in dist_info)

    return wheel_files


@contextlib.contextmanager
def clean_repository(complete_file_list: list[str]) -> t.Generator[str, None, None]:
    """Copy the files to a temporary directory and yield the path."""
    directories = sorted(set(os.path.dirname(path) for path in complete_file_list))
    directories.remove('')

    with tempfile.TemporaryDirectory() as temp_dir:
        for directory in directories:
            os.makedirs(os.path.join(temp_dir, directory))

        for path in complete_file_list:
            shutil.copy2(path, os.path.join(temp_dir, path), follow_symlinks=False)

        yield temp_dir


def build(source_dir: str, tmp_dir: str) -> tuple[pathlib.Path, pathlib.Path]:
    """Create a sdist and wheel."""
    create = subprocess.run(  # pylint: disable=subprocess-run-check
        [sys.executable, '-m', 'build', '--outdir', tmp_dir],
        stdin=subprocess.DEVNULL,
        capture_output=True,
        text=True,
        cwd=source_dir,
    )

    if create.returncode != 0:
        raise RuntimeError(f'build failed:\n{create.stderr}\n{create.stdout}')

    tmp_dir_files = list(pathlib.Path(tmp_dir).iterdir())

    if len(tmp_dir_files) != 2:
        raise RuntimeError(f'build resulted in {len(tmp_dir_files)} items instead of 2')

    sdist_path = [path for path in tmp_dir_files if path.suffix == '.gz'][0]
    wheel_path = [path for path in tmp_dir_files if path.suffix == '.whl'][0]

    return sdist_path, wheel_path


def list_sdist(path: pathlib.Path) -> list[str]:
    """Return a list of the files in the sdist."""
    item: tarfile.TarInfo

    with tarfile.open(path) as sdist:
        paths = ['/'.join(pathlib.Path(item.path).parts[1:]) for item in sdist.getmembers() if not item.isdir()]

    return paths


def list_wheel(path: pathlib.Path) -> list[str]:
    """Return a list of the files in the wheel."""
    with zipfile.ZipFile(path) as wheel:
        paths = [item.filename for item in wheel.filelist if not item.is_dir()]

    return paths


def filter_fuzzy_matches(missing: set[str], extra: set[str]) -> None:
    """
    Removes entries from `missing` and `extra` that share a common basename that also appears in `fuzzy_match_basenames`.
    Accounts for variable placement of non-runtime files by different versions of setuptools.
    """
    if not (missing or extra):
        return

    # calculate a set of basenames that appear in both missing and extra that are also marked as possibly needing fuzzy matching
    corresponding_fuzzy_basenames = {os.path.basename(p) for p in missing}.intersection(os.path.basename(p) for p in extra).intersection(fuzzy_match_basenames)

    # filter successfully fuzzy-matched entries from missing and extra
    missing.difference_update({p for p in missing if os.path.basename(p) in corresponding_fuzzy_basenames})
    extra.difference_update({p for p in extra if os.path.basename(p) in corresponding_fuzzy_basenames})


def check_files(source: str, expected: list[str], actual: list[str]) -> list[str]:
    """Verify the expected files exist and no extra files exist."""
    missing = set(expected) - set(actual)
    extra = set(actual) - set(expected)

    filter_fuzzy_matches(missing, extra)

    errors = [f'{path}: missing from {source}' for path in sorted(missing)] + [f'{path}: unexpected in {source}' for path in sorted(extra)]

    return errors


def main() -> None:
    """Main program entry point."""
    complete_file_list = sys.argv[1:] or sys.stdin.read().splitlines()

    python_version = '.'.join(map(str, sys.version_info[:2]))
    python_min = os.environ['ANSIBLE_TEST_MIN_PYTHON']
    python_max = os.environ['ANSIBLE_TEST_MAX_PYTHON']

    if python_version == python_min:
        use_upper_setuptools_version = False
    elif python_version == python_max:
        use_upper_setuptools_version = True
    else:
        raise RuntimeError(f'Python version {python_version} is neither the minimum {python_min} or the maximum {python_max}.')

    errors = check_build(complete_file_list, use_upper_setuptools_version)

    for error in errors:
        print(error)


def set_setuptools_version(repo_dir: str, use_upper_version: bool) -> str:
    pyproject_toml = pathlib.Path(repo_dir) / 'pyproject.toml'

    current = pyproject_toml.read_text()
    pattern = re.compile(r'^(?P<begin>requires = \["setuptools >= )(?P<lower>[^,]+)(?P<middle>, <= )(?P<upper>[^"]+)(?P<end>".*)$', re.MULTILINE)
    match = pattern.search(current)

    if not match:
        raise RuntimeError(f"Unable to find the 'requires' entry in: {pyproject_toml}")

    lower_version = match.group('lower')
    upper_version = match.group('upper')

    requested_version = upper_version if use_upper_version else lower_version

    updated = pattern.sub(fr'\g<begin>{requested_version}\g<middle>{requested_version}\g<end>', current)

    if current == updated:
        raise RuntimeError("Failed to set the setuptools version.")

    pyproject_toml.write_text(updated)

    return requested_version


def check_build(complete_file_list: list[str], use_upper_setuptools_version: bool) -> list[str]:
    errors: list[str] = []
    complete_file_list = list(complete_file_list)  # avoid mutation of input

    # Limit visible files to those reported by ansible-test.
    # This avoids including files which are not committed to git.
    with clean_repository(complete_file_list) as clean_repo_dir:
        setuptools_version = set_setuptools_version(clean_repo_dir, use_upper_setuptools_version)

        if __version__.endswith('.dev0'):
            # Make sure a changelog exists for this version when testing from devel.
            # When testing from a stable branch the changelog will already exist.
            major_minor_version = '.'.join(__version__.split('.')[:2])
            changelog_path = f'changelogs/CHANGELOG-v{major_minor_version}.rst'
            pathlib.Path(clean_repo_dir, changelog_path).touch()
            complete_file_list.append(changelog_path)

        expected_sdist_files = collect_sdist_files(complete_file_list)
        expected_wheel_files = collect_wheel_files(complete_file_list)

        with tempfile.TemporaryDirectory() as tmp_dir:
            sdist_path, wheel_path = build(clean_repo_dir, tmp_dir)

            actual_sdist_files = list_sdist(sdist_path)
            actual_wheel_files = list_wheel(wheel_path)

            errors.extend(check_files('sdist', expected_sdist_files, actual_sdist_files))
            errors.extend(check_files('wheel', expected_wheel_files, actual_wheel_files))

    errors = [f'{msg} ({setuptools_version})' for msg in errors]

    return errors


if __name__ == '__main__':
    main()
